import pytest
import scipy.stats as st
from numpy import concatenate
from numpy.random import RandomState
from numpy.testing import assert_allclose
from scipy.optimize import check_grad

from glimix_core.lmm import Kron2Sum


def test_lmm_kron2sum():
    random = RandomState(0)
    Y = random.randn(5, 3)
    A = random.randn(3, 3)
    A = A @ A.T
    F = random.randn(5, 2)
    G = random.randn(5, 4)
    lmm = Kron2Sum(Y, A, F, G)
    y = lmm._y

    m = lmm.mean.value()
    K = lmm.cov.value()
    assert_allclose(lmm.lml(), st.multivariate_normal(m, K).logpdf(y))

    lmm.mean.B = random.randn(2, 3)
    m = lmm.mean.value()
    assert_allclose(lmm.lml(), st.multivariate_normal(m, K).logpdf(y))

    lmm.cov.C0.Lu = random.randn(3)
    K = lmm.cov.value()
    assert_allclose(lmm.lml(), st.multivariate_normal(m, K).logpdf(y))

    lmm.cov.C1.Lu = random.randn(6)
    K = lmm.cov.value()
    assert_allclose(lmm.lml(), st.multivariate_normal(m, K).logpdf(y))


def test_lmm_kron2sum_gradient():
    random = RandomState(1)
    Y = random.randn(5, 3)
    A = random.randn(3, 3)
    A = A @ A.T
    F = random.randn(5, 2)
    G = random.randn(5, 4)
    lmm = Kron2Sum(Y, A, F, G)
    lmm.mean.B = random.randn(2, 3)
    lmm.cov.C0.Lu = random.randn(3)
    lmm.cov.C1.Lu = random.randn(6)

    def func(x):
        lmm.cov.C0.Lu = x[:3]
        lmm.cov.C1.Lu = x[3:9]
        lmm.mean.B = x[9:].reshape((2, 3), order="F")
        return lmm.lml()

    def grad(x):
        lmm.cov.C0.Lu = x[:3]
        lmm.cov.C1.Lu = x[3:9]
        lmm.mean.B = x[9:].reshape((2, 3), order="F")
        D = lmm.lml_gradient()
        return concatenate((D["C0.Lu"], D["C1.Lu"], D["M.vecB"]))

    assert_allclose(
        check_grad(func, grad, random.randn(15), epsilon=1e-8), 0, atol=1e-3
    )


def test_lmm_kron2sum_fit_ill_conditioned():
    random = RandomState(0)
    n = 30
    Y = random.randn(n, 3)
    A = random.randn(3, 3)
    A = A @ A.T
    F = random.randn(n, 2)
    G = random.randn(n, 4)
    lmm = Kron2Sum(Y, A, F, G)
    lml0 = lmm.lml()
    lmm.fit(verbose=False)
    lml1 = lmm.lml()
    assert_allclose([lml0, lml1], [-161.0096475502658, -122.97307222429187])
    grad = lmm.lml_gradient()
    vars = grad.keys()
    assert_allclose(concatenate([grad[var] for var in vars]), [0] * 15, atol=1e-3)


def test_lmm_kron2sum_fit_C1_well_cond():
    random = RandomState(0)
    Y = random.randn(5, 2)
    A = random.randn(2, 2)
    A = A @ A.T
    F = random.randn(5, 2)
    G = random.randn(5, 6)
    lmm = Kron2Sum(Y, A, F, G)
    lml0 = lmm.lml()
    lmm.fit(verbose=False)
    lml1 = lmm.lml()
    assert_allclose([lml0, lml1], [-21.459910525411757, -11.853021674440884])
    grad = lmm.lml_gradient()
    vars = grad.keys()
    assert_allclose(concatenate([grad[var] for var in vars]), [0] * 9, atol=1e-4)


def test_lmm_kron2sum_fit_C1_well_cond_C0_fullrank():
    random = RandomState(0)
    Y = random.randn(5, 2)
    A = random.randn(2, 2)
    A = A @ A.T
    F = random.randn(5, 2)
    G = random.randn(5, 6)
    lmm = Kron2Sum(Y, A, F, G, rank=2)
    lml0 = lmm.lml()
    lmm.fit(verbose=False)
    lml1 = lmm.lml()
    assert_allclose([lml0, lml1], [-22.44348328519218, -11.853021674565952])
    grad = lmm.lml_gradient()
    vars = grad.keys()
    assert_allclose(concatenate([grad[var] for var in vars]), [0] * 11, atol=1e-3)


# def test_lmm_kron2sum_fit_C1_well_cond_redutant_G():
#     random = RandomState(0)
#     Y = random.randn(5, 2)
#     A = random.randn(2, 2)
#     A = A @ A.T
#     F = random.randn(5, 2)
#     G = random.randn(5, 2)
#     G = concatenate((G, G), axis=1)
#     lmm = Kron2Sum(Y, A, F, G)
#     lml0 = lmm.lml()
#     lmm.fit(verbose=False)
#     lml1 = lmm.lml()
#     assert_allclose([lml0, lml1], [-19.905930526833977, 2.9779698820017453])
#     grad = lmm.lml_gradient()
#     vars = grad.keys()
#     assert_allclose(concatenate([grad[var] for var in vars]), [0] * 9, atol=1e-2)


def test_lmm_kron2sum_fit_C1_well_cond_redutant_F():
    random = RandomState(0)
    Y = random.randn(5, 2)
    A = random.randn(2, 2)
    A = A @ A.T
    F = random.randn(5, 2)
    F = concatenate((F, F), axis=1)
    G = random.randn(5, 2)
    lmm = Kron2Sum(Y, A, F, G)
    lml0 = lmm.lml()
    lmm.fit(verbose=False)
    lml1 = lmm.lml()
    assert_allclose([lml0, lml1], [-19.423945522925802, 2.9779698820395035])
    grad = lmm.lml_gradient()
    vars = grad.keys()
    assert_allclose(concatenate([grad[var] for var in vars]), [0] * 13, atol=1e-2)


def test_lmm_kron2sum_fit_C1_well_cond_redundant_Y():
    random = RandomState(0)
    Y = random.randn(5, 2)
    Y = concatenate((Y, Y), axis=1)
    A = random.randn(4, 4)
    A = A @ A.T
    F = random.randn(5, 2)
    G = random.randn(5, 2)
    with pytest.warns(UserWarning):
        lmm = Kron2Sum(Y, A, F, G)
    lml = lmm.lml()
    assert_allclose(lml, -43.906141655466485)



import warnings

from numpy import asfortranarray
from numpy.linalg import matrix_rank

from glimix_core._util import log2pi
from glimix_core.cov import Kron2SumCov
from glimix_core.mean import KronMean
from optimix import Function


class Kron2Sum(Function):
    """
    LMM for multiple traits fitted via maximum likelihood

    .. warning::

        This is a class in beta stage. Please, prefer using RKron2Sum instead.

    Let n, c, and p be the number of samples, covariates, and traits, respectively.
    The outcome variable Y is a n×p matrix distributed according to

        vec(Y) ~ N((A ⊗ F) vec(B), Cᵣ ⊗ GGᵗ + Cₙ ⊗ I).

    A and F are design matrices of dimensions p×p and n×c provided by the user,
    where F is the usual matrix of covariates commonly used in single-trait models.
    B is a p×c matrix of fixed-effect sizes.
    G is a n×r matrix provided by the user and I is a n×n identity matrices.
    Cᵣ and Cₙ are both symmetric matrices of dimensions p×p, for which Cₙ is
    guaranteed by our implementation to be of full rank.
    The parameters of this model are the matrices B, Cᵣ, and Cₙ.
    """

    def __init__(self, Y, A, F, G, rank=1):
        """
        Constructor.

        Parameters
        ----------
        Y : (n, p) array_like
            Outcome matrix.
        A : (n, n) array_like
            Trait-by-trait design matrix.
        F : (n, c) array_like
            Covariates design matrix.
        G : (n, r) array_like
            Matrix G from the GGᵗ term.
        rank : optional, int
            Maximum rank of matrix Cᵣ. Defaults to ``1``.
        """
        Y = asfortranarray(Y)
        yrank = matrix_rank(Y)
        if Y.shape[1] > yrank:
            warnings.warn(
                f"Y is not full column rank: rank(Y)={yrank}. "
                + "Convergence might be problematic.",
                UserWarning,
            )

        self._Y = Y
        self._y = Y.ravel(order="F")
        self._A = A
        self._F = F
        self._cov = Kron2SumCov(G, Y.shape[1], rank)
        self._mean = KronMean(A, F)
        composite = [("M", self._mean), ("C0", self._cov.C0), ("C1", self._cov.C1)]
        Function.__init__(self, "Kron2Sum", composite=composite)

    @property
    def mean(self):
        """
        Mean 𝐦 = (A ⊗ F) vec(B).

        Returns
        -------
        mean : KronMean
        """
        return self._mean

    @property
    def cov(self):
        """
        Covariance K = Cᵣ ⊗ GGᵗ + Cₙ ⊗ I.

        Returns
        -------
        covariance : Kron2SumCov
        """
        return self._cov

    @property
    def nsamples(self):
        """
        Number of samples, n.
        """
        return self._Y.shape[0]

    @property
    def ntraits(self):
        """
        Number of traits, p.
        """
        return self._Y.shape[1]

    @property
    def ncovariates(self):
        """
        Number of covariates, c.
        """
        return self._F.shape[1]

    def value(self):
        return self.lml()

    def gradient(self):
        return self.lml_gradient()

    def lml(self):
        r"""
        Log of the marginal likelihood.

        Let 𝐲 = vec(Y). The log of the marginal likelihood is given by

            log(p(𝐲)) = -n p log(2π) / 2 - log(\|K\|) / 2 - (𝐲-𝐦)ᵗ K⁻¹ (𝐲-𝐦) / 2.

        Returns
        -------
        lml : float
            Log of the marginal likelihood.
        """
        np = self.nsamples * self.ntraits
        lml = -np * log2pi - self._cov.logdet()

        m = self._mean.value()
        d = self._y - m
        dKid = d @ self._cov.solve(d)
        lml -= dKid

        return lml / 2

    def lml_gradient(self):
        """
        Gradient of the log of the marginal likelihood.

        Let 𝐲 = vec(Y) and 𝕂 = K⁻¹∂(K)K⁻¹. The gradient is given by

            2⋅∂log(p(𝐲)) = -tr(K⁻¹∂K) + 𝐲ᵗ𝕂𝐲 + (𝐦-2⋅𝐲)ᵗ𝕂𝐦 - 2⋅(𝐲-𝐦)ᵗK⁻¹∂(𝐦).

        Returns
        -------
        M.vecB : ndarray
            Gradient of the log of the marginal likelihood over vec(B).
        C0.Lu : ndarray
            Gradient of the log of the marginal likelihood over Cᵣ parameters.
        C1.Lu : ndarray
            Gradient of the log of the marginal likelihood over Cₙ parameters.
        """
        ld_grad = self._cov.logdet_gradient()
        dK = {n: g.transpose([2, 0, 1]) for (n, g) in self._cov.gradient().items()}
        Kiy = self._cov.solve(self._y)
        m = self._mean.value()
        Kim = self._cov.solve(m)
        grad = {}
        dm = self._mean.gradient()["vecB"]
        grad["M.vecB"] = dm.T @ Kiy - dm.T @ Kim
        for var in ["C0.Lu", "C1.Lu"]:
            grad[var] = -ld_grad[var]
            grad[var] += Kiy.T @ dK[var] @ Kiy
            grad[var] -= 2 * (Kim.T @ dK[var] @ Kiy)
            grad[var] += Kim.T @ dK[var] @ Kim
            grad[var] /= 2
        return grad

    def fit(self, verbose=True):
        """
        Maximise the marginal likelihood.

        Parameters
        ----------
        verbose : bool, optional
            ``True`` for progress output; ``False`` otherwise.
            Defaults to ``True``.
        """
        self._maximize(verbose=verbose)



from numpy import array
from numpy.testing import assert_allclose

from glimix_core.mean import LinearMean


def test_mean_linear():
    X = array([[5.1, 1.0], [2.1, -0.2]])

    mean = LinearMean(X)
    mean.effsizes = [1.0, -1.0]
    assert_allclose(mean.value(), X @ [1.0, -1.0])
    assert_allclose(mean.effsizes, [1.0, -1.0])
    assert_allclose(mean._check_grad(), 0, atol=1e-5)



from numpy import add

from optimix import Function


class SumMean(Function):
    """
    Sum mean function, 𝐟₀ + 𝐟₁ + ….

    The mathematical representation is

        𝐦 = 𝐟₀ + 𝐟₁ + …

    In other words, it is a sum of mean vectors.

    Example
    -------

    .. doctest::

        >>> from glimix_core.mean import OffsetMean, LinearMean, SumMean
        >>>
        >>> X = [[5.1, 1.0],
        ...      [2.1, -0.2]]
        >>>
        >>> mean0 = LinearMean(X)
        >>> mean0.effsizes = [-1.0, 0.5]
        >>>
        >>> mean1 = OffsetMean(2)
        >>> mean1.offset = 2.0
        >>>
        >>> mean = SumMean([mean0, mean1])
        >>>
        >>> print(mean.value())
        [-2.6 -0.2]
        >>> g = mean.gradient()
        >>> print(g["SumMean[0].effsizes"])
        [[ 5.1  1. ]
         [ 2.1 -0.2]]
        >>> print(g["SumMean[1].offset"])
        [1. 1.]
        >>> mean0.name = "A"
        >>> mean1.name = "B"
        >>> mean.name = "A+B"
        >>> print(mean)
        SumMean(means=...): A+B
          LinearMean(m=2): A
            effsizes: [-1.   0.5]
          OffsetMean(): B
            offset: 2.0
    """

    def __init__(self, means):
        """
        Constructor.

        Parameters
        ----------
        means : list
            List of mean functions.
        """
        self._means = [c for c in means]
        Function.__init__(self, "SumMean", composite=self._means)

    def value(self):
        """
        Sum of mean vectors, 𝐟₀ + 𝐟₁ + ….

        Returns
        -------
        𝐦 : ndarray
            𝐟₀ + 𝐟₁ + ….
        """
        return add.reduce([mean.value() for mean in self._means])

    def gradient(self):
        """
        Sum of mean function derivatives.

        Returns
        -------
        ∂𝐦 : dict
            ∂𝐟₀ + ∂𝐟₁ + ….
        """
        grad = {}
        for i, f in enumerate(self._means):
            for varname, g in f.gradient().items():
                grad[f"{self._name}[{i}].{varname}"] = g
        return grad

    def __str__(self):
        tname = type(self).__name__
        msg = "{}(means=...)".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        for m in self._means:
            spl = str(m).split("\n")
            msg = msg + "\n" + "\n".join(["  " + s for s in spl])
        return msg



from glimix_core.mean import OffsetMean
from numpy.testing import assert_allclose


def test_mean_offset():
    mean = OffsetMean(2)
    mean.offset = 1.5
    assert_allclose(mean.value(), [1.5, 1.5])
    assert_allclose(mean.offset, 1.5)
    assert_allclose(mean._check_grad(), 0, atol=1e-5)



from numpy import dot, empty, empty_like, sqrt, sum as npsum


class Posterior(object):
    r"""EP posterior.

    It is given by

    .. math::

        \mathbf z \sim \mathcal N\left(\Sigma (\tilde{\mathrm T}
          \tilde{\boldsymbol\mu} + \mathrm K^{-1}\mathbf m),
          (\tilde{\mathrm T} + \mathrm K^{-1})^{-1}\right).
    """

    def __init__(self, site):
        n = len(site.tau)
        self.tau = empty(n)
        self.eta = empty(n)
        self._site = site
        self._mean = None
        self._cov = None
        self._NxR_data = None
        self._RxN_data = None
        self._RxR_data = None
        self._L_cache = None
        self._LQt_cache = None
        self._QS_cache = None
        self._AQ_cache = None
        self._TAQ_cache = None
        self._QSQtATQLQtA_cache = None

    @property
    def _NxR(self):
        if self._NxR_data is None:
            Q = self._cov["QS"][0][0]
            self._NxR_data = empty_like(Q)
        return self._NxR_data

    @property
    def _RxN(self):
        if self._RxN_data is None:
            Q = self._cov["QS"][0][0]
            self._RxN_data = empty_like(Q.T)
        return self._RxN_data

    @property
    def _RxR(self):
        if self._RxR_data is None:
            Q = self._cov["QS"][0][0]
            r = Q.shape[1]
            self._RxR_data = empty((r, r))
        return self._RxR_data

    def _flush_cache(self):
        self._L_cache = None
        self._LQt_cache = None
        self._QS_cache = None
        self._AQ_cache = None
        self._TAQ_cache = None
        self._QSQtATQLQtA_cache = None

    def _initialize(self):
        r"""Initialize the mean and covariance of the posterior.

        Given that :math:`\tilde{\mathrm T}` is a matrix of zeros right before
        the first EP iteration, we have

        .. math::

            \boldsymbol\mu = \mathrm K^{-1} \mathbf m ~\text{ and }~
            \Sigma = \mathrm K

        as the initial posterior mean and covariance.
        """
        if self._mean is None or self._cov is None:
            return

        Q = self._cov["QS"][0][0]
        S = self._cov["QS"][1]

        if S.size > 0:
            self.tau[:] = 1 / npsum((Q * sqrt(S)) ** 2, axis=1)
        else:
            self.tau[:] = 0.0
        self.eta[:] = self._mean
        self.eta[:] *= self.tau

    @property
    def mean(self):
        return self._mean

    @mean.setter
    def mean(self, v):
        self._initialize()
        self._mean = v

    @property
    def cov(self):
        return self._cov

    @cov.setter
    def cov(self, v):
        self._initialize()
        self._cov = v

    def L(self):
        r"""Cholesky decomposition of :math:`\mathrm B`.

        .. math::

            \mathrm B = \mathrm Q^{\intercal}\tilde{\mathrm{T}}\mathrm Q
                + \mathrm{S}^{-1}
        """
        from scipy.linalg import cho_factor
        from numpy_sugar.linalg import ddot, sum2diag

        if self._L_cache is not None:
            return self._L_cache

        Q = self._cov["QS"][0][0]
        S = self._cov["QS"][1]
        B = dot(Q.T, ddot(self._site.tau, Q, left=True))
        sum2diag(B, 1.0 / S, out=B)
        self._L_cache = cho_factor(B, lower=True)[0]
        return self._L_cache

    def LQt(self):
        from numpy_sugar.linalg import cho_solve

        if self._LQt_cache is not None:
            return self._LQt_cache

        L = self.L()
        Q = self._cov["QS"][0][0]

        self._LQt_cache = cho_solve(L, Q.T)
        return self._LQt_cache

    def QS(self):
        from numpy_sugar.linalg import ddot

        if self._QS_cache is not None:
            return self._QS_cache

        Q = self._cov["QS"][0][0]
        S = self._cov["QS"][1]

        self._QS_cache = ddot(Q, S)
        return self._QS_cache

    def update(self):
        from numpy_sugar.linalg import ddot, dotd

        self._flush_cache()

        Q = self._cov["QS"][0][0]

        K = dot(self.QS(), Q.T)

        BiQt = self.LQt()
        TK = ddot(self._site.tau, K, left=True)
        BiQtTK = dot(BiQt, TK)

        self.tau[:] = K.diagonal()
        self.tau -= dotd(Q, BiQtTK)
        self.tau[:] = 1 / self.tau

        if not all(self.tau >= 0.0):
            raise RuntimeError("'tau' has to be non-negative.")

        self.eta[:] = dot(K, self._site.eta)
        self.eta[:] += self._mean
        self.eta[:] -= dot(Q, dot(BiQtTK, self._site.eta))
        self.eta[:] -= dot(Q, dot(BiQt, self._site.tau * self._mean))

        self.eta *= self.tau



import re
from os.path import join

from setuptools import find_packages


def get():
    pkgnames = find_packages()
    if len(pkgnames) == 0:
        return "unknown"
    pkgname = pkgnames[0]
    content = open(join(pkgname, "__init__.py")).read()
    c = re.compile(r"__version__ *= *('[^']+'|\"[^\"]+\")")
    m = c.search(content)
    if m is None:
        return "unknown"
    return m.groups()[0][1:-1]



from numpy.random import RandomState
from numpy.testing import assert_allclose

from glimix_core.cov import GivenCov


def test_givencov():
    K = RandomState(0).randn(5, 5)
    cov = GivenCov(K)
    assert_allclose(cov.value(), K, rtol=1e-4)
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    assert_allclose(cov.scale, 1.0)
    cov.scale = 1.5
    assert_allclose(cov.scale, 1.5)
    assert_allclose(cov.value(), 1.5 * K, rtol=1e-4)
    assert_allclose(cov._check_grad(), 0, atol=1e-5)



from numpy import asarray, ascontiguousarray, zeros

from optimix import Function, Vector


class LinearMean(Function):
    """
    Linear mean function, X𝜶.

    It defines X𝜶, for which X is a n×m matrix provided by the user and 𝜶 is a vector
    of size m.

    Example
    -------

    .. doctest::

        >>> from glimix_core.mean import LinearMean
        >>>
        >>> mean = LinearMean([[1.5, 0.2], [0.5, 0.4]])
        >>> mean.effsizes = [1.0, -1.0]
        >>> print(mean.value())
        [1.3 0.1]
        >>> print(mean.gradient()["effsizes"])
        [[1.5 0.2]
         [0.5 0.4]]
        >>> mean.name = "𝐦"
        >>> print(mean)
        LinearMean(m=2): 𝐦
          effsizes: [ 1. -1.]
    """

    def __init__(self, X):
        """
        Constructor.

        Parameters
        ----------
        X : array_like
            Covariates X, from X𝜶.
        """
        X = asarray(X, float)
        m = X.shape[1]
        self._effsizes = Vector(zeros(m))
        self._effsizes.bounds = [(-200.0, +200)] * m
        self._X = X
        Function.__init__(self, "LinearMean", effsizes=self._effsizes)

    @property
    def X(self):
        """
        An n×m matrix of covariates.
        """
        return self._X

    def value(self):
        """
        Linear mean function.

        Returns
        -------
        𝐦 : (n,) ndarray
            X𝜶.
        """
        return self._X @ self._effsizes

    def gradient(self):
        """
        Gradient of the linear mean function over the effect sizes.

        Returns
        -------
        effsizes : (n, m) ndarray
            X.
        """
        return dict(effsizes=self._X)

    @property
    def effsizes(self):
        """
        Effect-sizes parameter, 𝜶, of size m.
        """
        return self._effsizes.value

    @effsizes.setter
    def effsizes(self, v):
        self._effsizes.value = ascontiguousarray(v)

    def __str__(self):
        tname = type(self).__name__
        msg = "{}(m={})".format(tname, len(self.effsizes))
        if self.name is not None:
            msg += ": {}".format(self.name)
        msg += "\n"
        msg += "  effsizes: {}".format(self.effsizes)
        return msg



from numpy import full, ones

from optimix import Function, Scalar


class OffsetMean(Function):
    r"""
    Offset mean function, θ⋅𝟏.

    It represents a mean vector 𝐦 = θ⋅𝟏 of size n. The offset is given by the parameter
    θ.

    Example
    -------

    .. doctest::

        >>> from glimix_core.mean import OffsetMean
        >>>
        >>> mean = OffsetMean(3)
        >>> mean.offset = 2.0
        >>> print(mean.value())
        [2. 2. 2.]
        >>> print(mean.gradient())
        {'offset': array([1., 1., 1.])}
        >>> mean.name = "𝐦"
        >>> print(mean)
        OffsetMean(): 𝐦
          offset: 2.0
    """

    def __init__(self, n):
        """
        Constructor.

        Parameters
        ----------
        n : int
            Size of the 𝟏 array.
        """
        self._offset = Scalar(0.0)
        self._offset.bounds = (-200.0, +200)
        self._n = n
        Function.__init__(self, "OffsetMean", offset=self._offset)

    def fix_offset(self):
        """
        Prevent θ update during optimization.
        """
        self._fix("offset")

    def unfix_offset(self):
        """
        Enable θ update during optimization.
        """
        self._unfix("offset")

    def value(self):
        """
        Offset mean.

        Returns
        -------
        𝐦 : (n,) ndarray
            θ⋅𝟏.
        """
        return full(self._n, self._offset.value)

    def gradient(self):
        """
        Gradient of the offset function.

        Returns
        -------
        offset : (n,) ndarray
            Vector 𝟏.
        """
        return dict(offset=ones(self._n))

    @property
    def offset(self):
        """
        Offset parameter.
        """
        return self._offset.value

    @offset.setter
    def offset(self, v):
        self._offset.value = v

    def __str__(self):
        tname = type(self).__name__
        msg = "{}()".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        msg += "\n"
        msg += "  offset: {}".format(self.offset)
        return msg



"""
Covariance functions.
"""
from ._eye import EyeCov
from ._free import FreeFormCov
from ._given import GivenCov
from ._kron2sum import Kron2SumCov
from ._linear import LinearCov
from ._lrfree import LRFreeFormCov
from ._sum import SumCov

__all__ = [
    "EyeCov",
    "FreeFormCov",
    "GivenCov",
    "LinearCov",
    "SumCov",
    "LRFreeFormCov",
    "Kron2SumCov",
]



from numpy import exp, log

from optimix import Function, Scalar

from .._util import format_function


class GivenCov(Function):
    """
    Given covariance function, K = s⋅K₀.

    The covariance matrix is the provided matrix K₀ scaled by s: K = s⋅K₀.

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import GivenCov
        >>> from numpy import dot
        >>> from numpy.random import RandomState
        >>>
        >>> G = RandomState(0).randn(5, 3)
        >>> K0 = dot(G, G.T)
        >>> cov = GivenCov(K0)
        >>> cov.scale = 1.3
        >>> cov.name = "K"
        >>> print(cov)
        GivenCov(K0=...): K
          scale: 1.3
    """

    def __init__(self, K0):
        """
        Constructor.

        Parameters
        ----------
        K0 : array_like
            A semi-definite positive matrix.
        """
        self._logscale = Scalar(0.0)
        Function.__init__(self, "GivenCov", logscale=self._logscale)
        self._logscale.bounds = (-20.0, +10)
        self._K0 = K0

    @property
    def scale(self):
        """
        Scale parameter, s.
        """
        return exp(self._logscale.value)

    @scale.setter
    def scale(self, scale):
        from numpy_sugar import epsilon

        scale = max(scale, epsilon.tiny)
        self._logscale.value = log(scale)

    def value(self):
        """
        Covariance matrix, s⋅K₀.

        Returns
        -------
        K : ndarray
            s⋅K₀.
        """
        return self.scale * self._K0

    def gradient(self):
        """
        Derivative of the covariance matrix over log(s).

        Returns
        -------
        logscale : float
            s⋅K₀.
        """
        return dict(logscale=self.scale * self._K0)

    def __str__(self):
        return format_function(self, {"K0": "..."}, [("scale", self.scale)])



"""Random sampler.

GGPSampler       Sample from a Generalised Gaussian Process.
GPSampler        Sample from a Gaussian Process.
bernoulli_sample Bernoulli likelihood sampling.
binomial_sample  Binomial likelihood sampling.
poisson_sample   Poisson likelihood sampling.
"""

from ._ggp import GGPSampler
from ._gp import GPSampler
from ._canonical import bernoulli_sample, binomial_sample, poisson_sample

__all__ = [
    "GGPSampler",
    "GPSampler",
    "bernoulli_sample",
    "binomial_sample",
    "poisson_sample",
]



from numpy import asarray, exp, log


def _value_doc(func):
    func.__doc__ = r"""Link function evaluated at the given points.

        Parameters
        ----------
        x : array_like
            Array of points.

        Returns
        -------
        :class:`numpy.ndarray`
            Link function values.
        """
    return func


def _inv_doc(func):
    func.__doc__ = r"""Inverse of the link function evaluated at the given points.

        Parameters
        ----------
        x : array_like
            Array of points.

        Returns
        -------
        :class:`numpy.ndarray`
            Inverse of the link function values.
        """
    return func


class IdentityLink(object):
    r"""Identity link function.

    Mathematically,

    .. math::

        g(x) = x.
    """

    @_value_doc
    def value(self, x):
        return asarray(x, float)

    @_inv_doc
    def inv(self, x):
        return asarray(x, float)


class LogitLink(object):
    r"""Logit link function.

    Mathematically,

    .. math::

        g(x) = \log(x/(1 - x)).
    """

    @_value_doc
    def value(self, x):
        return asarray(log(x / (1 - x)), float)

    @_inv_doc
    def inv(self, x):
        return asarray(1 / (1 + exp(-x)), float)


class ProbitLink(object):
    r"""Probit link function.

    Mathematically,

    .. math::

        g(x) = \Phi^{-1}(x).
    """

    @_value_doc
    def value(self, x):
        return asarray(_normal_icdf(asarray(x, float)), float)

    @_inv_doc
    def inv(self, x):
        return asarray(_normal_cdf(asarray(x, float)), float)


class LogLink(object):
    r"""Log link function.

    Mathematically,

    .. math::

        g(x) = \log(x).
    """

    @_value_doc
    def value(self, x):
        return asarray(log(x), float)

    @_inv_doc
    def inv(self, x):
        return asarray(exp(x), float)


def _normal_cdf(x):
    import scipy.stats as st

    return st.norm.cdf(x)


def _normal_icdf(x):
    import scipy.stats as st

    return st.norm.isf(1 - x)



"""Link functions.

IdentityLink  Identity link function.
LogitLink     Logit link function.
LogLink       Log link function.
ProbitLink    Probit link function.
"""
from ._link import IdentityLink, LogitLink, LogLink, ProbitLink

__all__ = ["IdentityLink", "LogitLink", "LogLink", "ProbitLink"]



from copy import copy

from liknorm import LikNormMachine
from numpy import asarray, dot, exp
from numpy.linalg import solve

from glimix_core._ep import EPLinearKernel

from ._glmm import GLMM


class GLMMExpFam(GLMM):
    r"""Generalised Linear Gaussian Processes implementation.

    It implements inference over GLMMs via the Expectation Propagation [Min01]_
    algorithm.
    It currently supports the ``"Bernoulli"``, ``"Probit"``, ``"Binomial"``, and
    ``"Poisson"`` likelihoods. (For heterogeneous Normal likelihood, please refer to
    :class:`glimix_core.glmm.GLMMNormal` for a closed-form inference.)

    Parameters
    ----------
    y : array_like
        Outcome variable.
    lik : tuple
        Likelihood definition. The first item is one of the following likelihood names:
        ``"Bernoulli"``, ``"Binomial"``, ``"Normal"``, and ``"Poisson"``. For
        `Binomial`, the second item is an array of outcomes.
    X : array_like
        Covariates.
    QS : tuple
        Economic eigendecomposition.

    Example
    -------

    .. doctest::

        >>> from numpy import dot, sqrt, zeros
        >>> from numpy.random import RandomState
        >>>
        >>> from numpy_sugar.linalg import economic_qs
        >>>
        >>> from glimix_core.glmm import GLMMExpFam
        >>>
        >>> random = RandomState(0)
        >>> nsamples = 10
        >>>
        >>> X = random.randn(nsamples, 2)
        >>> G = random.randn(nsamples, 100)
        >>> K = dot(G, G.T)
        >>> ntrials = random.randint(1, 100, nsamples)
        >>> z = dot(G, random.randn(100)) / sqrt(100)
        >>>
        >>> successes = zeros(len(ntrials), int)
        >>> for i in range(len(ntrials)):
        ...    successes[i] = sum(z[i] + 0.2 * random.randn(ntrials[i]) > 0)
        >>>
        >>> QS = economic_qs(K)
        >>>
        >>> glmm = GLMMExpFam(successes, ('binomial', ntrials), X, QS)
        >>> print('Before: %.2f' % glmm.lml())
        Before: -16.40
        >>> glmm.fit(verbose=False)
        >>> print('After: %.2f' % glmm.lml())
        After: -13.43
    """

    def __init__(
        self, y, lik, X, QS=None, n_int=1000, rtol=1.49e-08 * 1000, atol=1.49e-08
    ):
        GLMM.__init__(self, y, lik, X, QS)

        self._ep = EPLinearKernel(self._X.shape[0], rtol=rtol, atol=atol)
        self._ep.set_compute_moments(self.compute_moments)
        self._machine = LikNormMachine(self._lik[0], n_int)
        self.update_approx = True

        self._variables.get("beta").listen(self.set_update_approx)
        self._variables.get("logscale").listen(self.set_update_approx)
        self._variables.get("logitdelta").listen(self.set_update_approx)

    def __copy__(self):
        gef = GLMMExpFam(self._y, self._lik, self._X, self._QS)
        gef.__dict__["_ep"] = copy(self._ep)
        gef.__dict__["_ep"].set_compute_moments(gef.compute_moments)
        gef.update_approx = self.update_approx

        GLMM._copy_to(self, gef)

        return gef

    def __del__(self):
        if hasattr(self, "_machine"):
            self._machine.finish()

    def _update_approx(self):
        if not self.update_approx:
            return

        self._ep.set_prior(self.mean(), self.covariance())
        self.update_approx = False

    @property
    def beta(self):
        return GLMM.beta.fget(self)

    @beta.setter
    def beta(self, v):
        GLMM.beta.fset(self, v)
        self.set_update_approx()

    def compute_moments(self, eta, tau, moments):
        y = (self._y,) + self._lik[1:]
        self._machine.moments(y, eta, tau, moments)

    def covariance(self):
        return dict(QS=self._QS, scale=self.scale, delta=self.delta)

    def fit(self, verbose=True, factr=1e5, pgtol=1e-7):
        self._ep.verbose = verbose
        super(GLMMExpFam, self).fit(verbose=verbose, factr=factr, pgtol=pgtol)
        self._ep.verbose = False

    def fix(self, var_name):
        GLMM.fix(self, var_name)
        self.set_update_approx()

    def gradient(self):
        r"""Gradient of the log of the marginal likelihood.

        Returns
        -------
        dict
            Map between variables to their gradient values.
        """
        self._update_approx()

        g = self._ep.lml_derivatives(self._X)
        ed = exp(-self.logitdelta)
        es = exp(self.logscale)

        grad = dict()
        grad["logitdelta"] = g["delta"] * (ed / (1 + ed)) / (1 + ed)
        grad["logscale"] = g["scale"] * es
        grad["beta"] = g["mean"]

        return grad

    @property
    def logitdelta(self):
        return GLMM.logitdelta.fget(self)

    @logitdelta.setter
    def logitdelta(self, v):
        GLMM.logitdelta.fset(self, v)
        self.set_update_approx()

    @property
    def logscale(self):
        return GLMM.logscale.fget(self)

    @logscale.setter
    def logscale(self, v):
        GLMM.logscale.fset(self, v)
        self.set_update_approx()

    def set_update_approx(self, _=None):
        self.update_approx = True

    def set_variable_bounds(self, var_name, bounds):
        GLMM.set_variable_bounds(self, var_name, bounds)
        self.set_update_approx()

    @property
    def site(self):
        return self._ep.site

    def unfix(self, var_name):
        GLMM.unfix(self, var_name)
        self.set_update_approx()

    def value(self):
        self._update_approx()
        return self._ep.lml()

    def predictive_mean(self, Xstar, ks, kss):
        mstar = self.mean_star(Xstar)
        ks = self.covariance_star(ks)
        m = self.mean()
        eta = self._ep.posterior.eta
        tau = self._ep.posterior.tau
        mu = eta / tau
        K = GLMM.covariance(self)
        return mstar + dot(ks, solve(K, mu - m))

    def predictive_covariance(self, Xstar, ks, kss):
        from numpy_sugar.linalg import sum2diag

        kss = self.variance_star(kss)
        ks = self.covariance_star(ks)
        tau = self._ep.posterior.tau
        K = GLMM.covariance(self)
        KT = sum2diag(K, 1 / tau)
        ktk = solve(KT, ks.T)
        b = []
        for i in range(len(kss)):
            b += [dot(ks[i, :], ktk[:, i])]
        b = asarray(b)
        return kss - b



import pytest
from numpy import (
    arange,
    asarray,
    ascontiguousarray,
    corrcoef,
    dot,
    eye,
    ones,
    sqrt,
    zeros,
)
from numpy.random import RandomState
from numpy.testing import assert_, assert_allclose

from glimix_core.example import linear_eye_cov
from glimix_core.glmm import GLMMExpFam, GLMMNormal
from glimix_core.random import bernoulli_sample
from numpy_sugar.linalg import economic_qs, economic_qs_linear

ATOL = 1e-3
RTOL = 1e-3


def test_glmmexpfam_layout():
    y = asarray([1.0, 0.5])
    X = asarray([[0.5, 1.0]])
    K = asarray([[1.0, 0.0], [0.0, 1.0]])
    QS = economic_qs(K)

    with pytest.raises(ValueError):
        GLMMExpFam(y, "poisson", X, QS=QS)

    y = asarray([1.0])
    with pytest.raises(ValueError):
        GLMMExpFam(y, "poisson", X, QS=QS)


def test_glmmexpfam_copy():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = linear_eye_cov().value()
    z = random.multivariate_normal(0.2 * ones(nsamples), K)
    QS = economic_qs(K)

    ntri = random.randint(1, 30, nsamples)
    nsuc = zeros(nsamples, dtype=int)
    for (i, ni) in enumerate(ntri):
        nsuc[i] += sum(z[i] + 0.2 * random.randn(ni) > 0)

    ntri = ascontiguousarray(ntri)
    glmm0 = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)

    assert_allclose(glmm0.lml(), -29.10216812909928, atol=ATOL, rtol=RTOL)
    glmm0.fit(verbose=False)

    v = -19.575736562427252
    assert_allclose(glmm0.lml(), v)

    glmm1 = glmm0.copy()
    assert_allclose(glmm1.lml(), v)

    glmm1.scale = 0.92
    assert_allclose(glmm0.lml(), v, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm1.lml(), -30.832831740038056, atol=ATOL, rtol=RTOL)

    glmm0.fit(verbose=False)
    glmm1.fit(verbose=False)

    v = -19.575736562378573
    assert_allclose(glmm0.lml(), v)
    assert_allclose(glmm1.lml(), v)


def test_glmmexpfam_precise():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.randint(1, 30, nsamples)
    nsuc = [random.randint(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ["binomial", ntri], X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.scale = 1.0
    assert_allclose(glmm.lml(), -44.74191041468836, atol=ATOL, rtol=RTOL)
    glmm.scale = 2.0
    assert_allclose(glmm.lml(), -36.19907331929086, atol=ATOL, rtol=RTOL)
    glmm.scale = 3.0
    assert_allclose(glmm.lml(), -33.02139830387104, atol=ATOL, rtol=RTOL)
    glmm.scale = 4.0
    assert_allclose(glmm.lml(), -31.42553401678996, atol=ATOL, rtol=RTOL)
    glmm.scale = 5.0
    assert_allclose(glmm.lml(), -30.507029479473243, atol=ATOL, rtol=RTOL)
    glmm.scale = 6.0
    assert_allclose(glmm.lml(), -29.937569702301232, atol=ATOL, rtol=RTOL)
    glmm.delta = 0.1
    assert_allclose(glmm.lml(), -30.09977907145003, atol=ATOL, rtol=RTOL)

    assert_allclose(glmm._check_grad(), 0, atol=1e-3, rtol=RTOL)


def test_glmmexpfam_glmmnormal_get_fast_scanner():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    eta = random.randn(nsamples)
    tau = 10 * random.rand(nsamples)

    glmm = GLMMNormal(eta, tau, X, QS)
    glmm.fit(verbose=False)
    want = [-0.08228058, -0.03910674, 0.04226152, -0.05893827, 0.01718722]
    assert_allclose(glmm.beta, want, atol=1e-3, rtol=1e-3)
    assert_allclose(0.001, glmm.scale, atol=1e-3, rtol=1e-3)
    assert_allclose(0.999999994119, glmm.delta, atol=1e-3, rtol=1e-3)

    scanner = glmm.get_fast_scanner()
    scanner.set_scale(1.0)
    lmls, effect_sizes = scanner.fast_scan(X, verbose=False)
    want = [-4.75845, -4.75845, -4.75845, -4.75845, -4.75845]
    assert_allclose(lmls, want, atol=1e-2, rtol=1e-2)

    want = [-0.04114, -0.019553, 0.021131, -0.029469, 0.008593]
    assert_allclose(effect_sizes, want, atol=1e-2, rtol=1e-2)


def test_glmmexpfam_delta0():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.randint(1, 30, nsamples)
    nsuc = [random.randint(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.delta = 0

    assert_allclose(glmm.lml(), -43.154282363439364, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm._check_grad(step=2e-5), 0, atol=1e-2)


def test_glmmexpfam_delta1():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.randint(1, 30, nsamples)
    nsuc = [random.randint(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.delta = 1

    assert_allclose(glmm.lml(), -47.09677870648636, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm._check_grad(), 0, atol=1e-4)


def test_glmmexpfam_wrong_qs():
    random = RandomState(0)
    X = random.randn(10, 15)
    linear_eye_cov().value()
    QS = [0, 1]

    ntri = random.randint(1, 30, 10)
    nsuc = [random.randint(0, i) for i in ntri]

    with pytest.raises(ValueError):
        GLMMExpFam((nsuc, ntri), "binomial", X, QS)


def test_glmmexpfam_optimize():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = linear_eye_cov().value()
    z = random.multivariate_normal(0.2 * ones(nsamples), K)
    QS = economic_qs(K)

    ntri = random.randint(1, 30, nsamples)
    nsuc = zeros(nsamples, dtype=int)
    for (i, ni) in enumerate(ntri):
        nsuc[i] += sum(z[i] + 0.2 * random.randn(ni) > 0)

    ntri = ascontiguousarray(ntri)
    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)

    assert_allclose(glmm.lml(), -29.102168129099287, atol=ATOL, rtol=RTOL)
    glmm.fix("beta")
    glmm.fix("scale")

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -27.635788105778012, atol=ATOL, rtol=RTOL)

    glmm.unfix("beta")
    glmm.unfix("scale")

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -19.68486269551159, atol=ATOL, rtol=RTOL)


def test_glmmexpfam_optimize_low_rank():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = dot(X, X.T)
    z = dot(X, 0.2 * random.randn(5))
    QS = economic_qs(K)

    ntri = random.randint(1, 30, nsamples)
    nsuc = zeros(nsamples, dtype=int)
    for (i, ni) in enumerate(ntri):
        nsuc[i] += sum(z[i] + 0.2 * random.randn(ni) > 0)

    ntri = ascontiguousarray(ntri)
    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)

    assert_allclose(glmm.lml(), -18.60476792256323, atol=ATOL, rtol=RTOL)
    glmm.fit(verbose=False)
    assert_allclose(glmm.lml(), -7.800621320491801, atol=ATOL, rtol=RTOL)


def test_glmmexpfam_bernoulli_problematic():
    random = RandomState(1)
    N = 30
    G = random.randn(N, N + 50)
    y = bernoulli_sample(0.0, G, random_state=random)

    G = ascontiguousarray(G, dtype=float)
    _stdnorm(G, 0, out=G)
    G /= sqrt(G.shape[1])

    QS = economic_qs_linear(G)
    S0 = QS[1]
    S0 /= S0.mean()

    X = ones((len(y), 1))
    model = GLMMExpFam(y, "bernoulli", X, QS=(QS[0], QS[1]))
    model.delta = 0
    model.fix("delta")
    model.fit(verbose=False)
    assert_allclose(model.lml(), -20.727007958026853, atol=ATOL, rtol=RTOL)
    assert_allclose(model.delta, 0, atol=1e-3)
    assert_allclose(model.scale, 0.879915823030081, atol=ATOL, rtol=RTOL)
    assert_allclose(model.beta, [-0.00247856564728], atol=ATOL, rtol=RTOL)


def test_glmmexpfam_bernoulli_probit_problematic():
    random = RandomState(1)
    N = 30
    G = random.randn(N, N + 50)
    y = bernoulli_sample(0.0, G, random_state=random)

    G = ascontiguousarray(G, dtype=float)
    _stdnorm(G, 0, out=G)
    G /= sqrt(G.shape[1])

    QS = economic_qs_linear(G)
    S0 = QS[1]
    S0 /= S0.mean()

    X = ones((len(y), 1))
    model = GLMMExpFam(y, "probit", X, QS=(QS[0], QS[1]))
    model.delta = 0
    model.fix("delta")
    model.fit(verbose=False)
    assert_allclose(model.lml(), -20.725623168378615, atol=ATOL, rtol=RTOL)
    assert_allclose(model.delta, 0.0001220703125, atol=1e-3)
    assert_allclose(model.scale, 0.33022865011938707, atol=ATOL, rtol=RTOL)
    assert_allclose(model.beta, [-0.002617161564786044], atol=ATOL, rtol=RTOL)

    h20 = model.scale * (1 - model.delta) / (model.scale + 1)

    model.unfix("delta")
    model.delta = 0.5
    model.scale = 1.0
    model.fit(verbose=False)

    assert_allclose(model.lml(), -20.725623168378522, atol=ATOL, rtol=RTOL)
    assert_allclose(model.delta, 0.5017852859580029, atol=1e-3)
    assert_allclose(model.scale, 0.9928931515372, atol=ATOL, rtol=RTOL)
    assert_allclose(model.beta, [-0.003203427206253548], atol=ATOL, rtol=RTOL)

    h21 = model.scale * (1 - model.delta) / (model.scale + 1)

    assert_allclose(h20, h21, atol=ATOL, rtol=RTOL)


def test_glmmexpfam_bernoulli_probit_assure_delta_fixed():
    random = RandomState(1)
    N = 10
    G = random.randn(N, N + 50)
    y = bernoulli_sample(0.0, G, random_state=random)

    G = ascontiguousarray(G, dtype=float)
    _stdnorm(G, 0, out=G)
    G /= sqrt(G.shape[1])

    QS = economic_qs_linear(G)
    S0 = QS[1]
    S0 /= S0.mean()

    X = ones((len(y), 1))
    model = GLMMExpFam(y, "probit", X, QS=(QS[0], QS[1]))
    model.fit(verbose=False)

    assert_allclose(model.lml(), -6.108751595773174, rtol=RTOL)
    assert_allclose(model.delta, 1.4901161193847673e-08, atol=1e-5)
    assert_(model._isfixed("logitdelta"))


def _stdnorm(X, axis=None, out=None):
    X = ascontiguousarray(X)
    if out is None:
        out = X.copy()

    m = out.mean(axis)
    s = out.std(axis)
    ok = s > 0

    out -= m

    if out.ndim == 1:
        if s > 0:
            out /= s
    else:
        out[..., ok] /= s[ok]

    return out


def test_glmmexpfam_binomial_pheno_list():
    random = RandomState(0)
    n = 10

    X = random.randn(n, 2)
    G = random.randn(n, 100)
    K = dot(G, G.T)
    ntrials = random.randint(1, 100, n)
    z = dot(G, random.randn(100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i in range(len(ntrials)):
        for _ in range(ntrials[i]):
            successes[i] += int(z[i] + 0.1 * random.randn() > 0)

    QS = economic_qs(K)
    glmm = GLMMExpFam(successes, ("binomial", ntrials), X, QS)
    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -11.438821585341866)


def test_glmmexpfam_binomial_large_ntrials():
    random = RandomState(0)
    n = 10

    X = random.randn(n, 2)
    G = random.randn(n, 100)
    K = dot(G, G.T)
    ntrials = random.randint(1, 100000, n)
    z = dot(G, random.randn(100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i in range(len(ntrials)):
        for _ in range(ntrials[i]):
            successes[i] += int(z[i] + 0.1 * random.randn() > 0)

    QS = economic_qs(K)
    glmm = GLMMExpFam(successes, ("binomial", ntrials), X, QS)
    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -43.067433588125446)


def test_glmmexpfam_scale_very_low():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.randint(1, 30, nsamples)
    nsuc = [random.randint(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.scale = 1e-3
    assert_allclose(glmm.lml(), -145.01170823743104, atol=ATOL, rtol=RTOL)

    assert_allclose(glmm._check_grad(), 0, atol=1e-2)


def test_glmmexpfam_scale_very_high():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.randint(1, 30, nsamples)
    nsuc = [random.randint(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.scale = 30.0
    assert_allclose(glmm.lml(), -29.632791380478736, atol=ATOL, rtol=RTOL)

    assert_allclose(glmm._check_grad(), 0, atol=1e-3)


def test_glmmexpfam_delta_one_zero():
    random = RandomState(1)
    n = 30
    X = random.randn(n, 6)
    K = dot(X, X.T)
    K /= K.diagonal().mean()
    QS = economic_qs(K)

    ntri = random.randint(1, 30, n)
    nsuc = [random.randint(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4, -0.2])

    glmm.delta = 0
    assert_allclose(glmm.lml(), -113.24570457063275)
    assert_allclose(glmm._check_grad(step=1e-4), 0, atol=1e-2)

    glmm.fit(verbose=False)
    assert_allclose(glmm.lml(), -98.21144899310399, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm.delta, 0, atol=ATOL, rtol=RTOL)

    glmm.delta = 1
    assert_allclose(glmm.lml(), -98.00058169240869, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm._check_grad(step=1e-4), 0, atol=1e-1)

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -72.82680948264196, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm.delta, 0.9999999850988439, atol=ATOL, rtol=RTOL)


def test_glmmexpfam_predict():

    random = RandomState(4)
    n = 100
    p = n + 1

    X = ones((n, 2))
    X[:, 1] = random.randn(n)

    G = random.randn(n, p)
    G /= G.std(0)
    G -= G.mean(0)
    G /= sqrt(p)
    K = dot(G, G.T)

    i = asarray(arange(0, n), int)
    si = random.choice(i, n, replace=False)
    ntest = int(n // 5)
    itrain = si[:-ntest]
    itest = si[-ntest:]

    Xtrain = X[itrain, :]
    Ktrain = K[itrain, :][:, itrain]

    Xtest = X[itest, :]

    beta = random.randn(2)
    z = random.multivariate_normal(dot(X, beta), 0.9 * K + 0.1 * eye(n))

    ntri = random.randint(1, 100, n)
    nsuc = zeros(n, dtype=int)
    for (i, ni) in enumerate(ntri):
        nsuc[i] += sum(z[i] + 0.2 * random.randn(ni) > 0)

    ntri = ascontiguousarray(ntri)

    QStrain = economic_qs(Ktrain)
    nsuc_train = ascontiguousarray(nsuc[itrain])
    ntri_train = ascontiguousarray(ntri[itrain])

    nsuc_test = ascontiguousarray(nsuc[itest])
    ntri_test = ascontiguousarray(ntri[itest])

    glmm = GLMMExpFam(nsuc_train, ("binomial", ntri_train), Xtrain, QStrain)
    glmm.fit(verbose=False)
    ks = K[itest, :][:, itrain]
    kss = asarray([K[i, i] for i in itest])
    pm = glmm.predictive_mean(Xtest, ks, kss)
    pk = glmm.predictive_covariance(Xtest, ks, kss)
    r = nsuc_test / ntri_test
    assert_(corrcoef([pm, r])[0, 1] > 0.8)
    assert_allclose(pk[0], 54.26335297363809)


def test_glmmexpfam_qs_none():
    nsamples = 10

    random = RandomState(0)
    X = random.randn(nsamples, 5)
    K = linear_eye_cov().value()
    z = random.multivariate_normal(0.2 * ones(nsamples), K)

    ntri = random.randint(1, 30, nsamples)
    nsuc = zeros(nsamples, dtype=int)
    for (i, ni) in enumerate(ntri):
        nsuc[i] += sum(z[i] + 0.2 * random.randn(ni) > 0)

    ntri = ascontiguousarray(ntri)
    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, None)

    assert_allclose(glmm.lml(), -38.30173374439622, atol=ATOL, rtol=RTOL)
    glmm.fix("beta")
    glmm.fix("scale")

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -32.03927471370041, atol=ATOL, rtol=RTOL)

    glmm.unfix("beta")
    glmm.unfix("scale")

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -19.575736561760586, atol=ATOL, rtol=RTOL)



from numpy.testing import assert_allclose

from glimix_core.link import LogitLink, LogLink, ProbitLink


def test_probit_link():
    link = ProbitLink()
    assert_allclose(link.value(link.inv(3.2)), 3.2)


def test_logit_link():
    link = LogitLink()
    assert_allclose(link.value(link.inv(3.2)), 3.2)


def test_loglink_link():
    link = LogLink()
    assert_allclose(link.value(link.inv(3.2)), 3.2)



from numpy.random import RandomState
from numpy.testing import assert_allclose, assert_equal

from glimix_core._util import assert_interface
from glimix_core.lmm import RKron2Sum


def test_lmm_rkron2sum():
    random = RandomState(0)
    n = 5
    Y = random.randn(n, 3)
    A = random.randn(3, 3)
    A = A @ A.T
    F = random.randn(n, 2)
    G = random.randn(n, 4)
    lmm = RKron2Sum(Y, A, F, G)

    assert_allclose(lmm.lml(), -16.580821931417656)
    assert_allclose(lmm._check_grad(step=1e-7), 0, atol=1e-4)
    assert_equal(lmm.nsamples, n)
    assert_equal(lmm.ntraits, 3)
    assert_equal(lmm.ncovariates, 2)

    n = 5
    Y = random.randn(n, 1)
    A = random.randn(1, 1)
    A = A @ A.T
    F = random.randn(n, 2)
    G = random.randn(n, 4)
    lmm = RKron2Sum(Y, A, F, G)
    lmm.name = "KronSum"

    assert_allclose(lmm.lml(), -4.582089407009583)
    assert_allclose(lmm._check_grad(step=1e-7), 0, atol=1e-4)
    assert_allclose(
        [lmm.mean.value()[0], lmm.mean.value()[1]],
        [0.0497438970225256, 0.5890598193072355],
    )

    assert_allclose(
        [
            lmm.cov.value()[0, 0],
            lmm.cov.value()[0, 1],
            lmm.cov.value()[1, 0],
            lmm.cov.value()[1, 1],
        ],
        [
            4.3712532668348185,
            -0.07239366121399138,
            -0.07239366121399138,
            2.7242131674614862,
        ],
    )

    assert_equal(lmm.nsamples, n)
    assert_equal(lmm.ntraits, 1)
    assert_equal(lmm.name, "KronSum")
    lmm.fit(verbose=False)
    grad = lmm.gradient()
    assert_allclose(grad["C0.Lu"], [0], atol=1e-4)
    assert_allclose(grad["C1.Lu"], [0], atol=1e-4)
    assert_allclose(lmm.lml(), -0.6930197958236421)


def test_lmm_rkron2sum_public_attrs():
    assert_interface(
        RKron2Sum,
        [
            "fit",
            "ntraits",
            "lml",
            "mean",
            "nsamples",
            "value",
            "cov",
            "ncovariates",
            "name",
            "gradient",
        ],
    )



"""
Linear mixed models.
"""
from ._kron2sum import Kron2Sum
from ._lmm import LMM
from ._mt_lmm import MTLMM
from ._rkron2sum import RKron2Sum
from ._scan import FastScanner

__all__ = ["LMM", "MTLMM", "FastScanner", "Kron2Sum", "RKron2Sum"]



from . import numbers
from ._array import unvec, vec
from .check import check_covariates, check_economic_qs, check_outcome
from .eigen import economic_qs_zeros
from .format import format_function
from .hsolve import hsolve
from ._assert import assert_interface

log2pi = 1.837877066409345339081937709124758839607238769531250

__all__ = [
    "check_covariates",
    "check_economic_qs",
    "check_outcome",
    "economic_qs_zeros",
    "hsolve",
    "numbers",
    "vec",
    "unvec",
    "format_function",
    "assert_interface",
]



from numpy import array
from numpy.testing import assert_allclose

from glimix_core.cov import FreeFormCov


def test_freeformcov():

    cov = FreeFormCov(3)
    L = array([[1.5, 0, 0], [2.5, 0.8, 0], [-0.3, 0.4, 2.1]], float)
    cov.L = L
    assert_allclose(cov.value(), L @ L.T, rtol=1e-4)
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    assert_allclose(cov.logdet(), 1.8486202109176444)

    cov = FreeFormCov(2)
    L = array([[1, 0], [2.5, 1]], float)
    cov.L = L
    assert_allclose(cov.value(), L @ L.T, rtol=1e-4)
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    assert_allclose(cov.logdet(), 0.00012292724603080174)

    cov = FreeFormCov(1)
    L = array([[0.5]], float)
    cov.L = L
    assert_allclose(cov.value(), L @ L.T, rtol=1e-4)
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    assert_allclose(cov.logdet(), -1.3862347582514014)



import numpy as np
from numpy.testing import assert_allclose

from glimix_core.cov import GivenCov, LinearCov, SumCov


def test_sumcov():
    random = np.random.RandomState(0)
    X = random.randn(3, 2)
    cov_left = LinearCov(X)

    K = random.randn(3, 3)
    cov_right = GivenCov(K)

    cov = SumCov([cov_left, cov_right])
    assert_allclose(cov.value(), cov_left.value() + cov_right.value())
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    cov_left.scale = 0.1
    assert_allclose(cov._check_grad(), 0, atol=1e-5)



def assert_interface(cls, req_attrs):
    attrs = dir(cls)
    private = set(a for a in attrs if a.startswith("_"))
    public = set(attrs) - private

    missing = public - set(req_attrs)
    if missing:
        msg = "The following public attributes exist but have not been asserted: "
        msg += "{}".format(", ".join(list(missing)))
        raise AssertionError(msg)

    missing = set(req_attrs) - public
    if missing:
        msg = "The following attributes have not been found: "
        msg += "{}".format(", ".join(list(missing)))
        raise AssertionError(msg)



import warnings

from numpy import all as npall, ascontiguousarray, clip, isfinite


def check_economic_qs(QS):
    if not isinstance(QS, tuple):
        raise ValueError("QS must be a tuple.")

    if not isinstance(QS[0], tuple):
        raise ValueError("QS[0] must be a tuple.")

    fmsg = "QS has non-finite values."

    if not all(npall(isfinite(Q)) for Q in QS[0]):
        raise ValueError(fmsg)

    if not npall(isfinite(QS[1])):
        raise ValueError(fmsg)

    return QS


def check_covariates(X):
    if not X.ndim == 2:
        raise ValueError("Covariates must be a bidimensional array.")

    if not npall(isfinite(X)):
        raise ValueError("Covariates must have finite values only.")

    return X


def check_outcome(y, lik):
    if not isinstance(lik, (list, tuple)):
        lik = (lik,)

    str_err = "The first item of ``lik`` has to be a string."
    if not isinstance(lik[0], str):
        raise ValueError(str_err)

    lik_name = lik[0].lower()

    y = ascontiguousarray(y, float)
    lik = lik[:1] + tuple(ascontiguousarray(i, float) for i in lik[1:])

    if not npall(isfinite(y)):
        raise ValueError("Outcome must be finite.")

    if lik_name == "poisson":
        return _check_poisson_outcome(y)

    if lik_name in ("binomial", "normal"):
        if len(lik) != 2:
            msg = "``lik`` must be a tuple of two elements for"
            msg += " {} likelihood.".format(lik_name[0].upper() + lik_name[1:])
            raise ValueError(msg)

    return y


def _check_poisson_outcome(y):
    poisson_lim = 25000

    if y.max() > poisson_lim:
        msg = "Output values of Poisson likelihood greater"
        msg += " than {lim} is set to {lim} before applying GLMM."
        warnings.warn(msg.format(lim=poisson_lim))
        y = clip(y, 0, poisson_lim)

    return y



from numpy import dot

from .posterior import Posterior


def _cho_factor(B):
    from scipy.linalg import cho_factor

    B = cho_factor(B, overwrite_a=True, lower=True, check_finite=False)[0]
    return B


class PosteriorLinearKernel(Posterior):
    """
    EP posterior.

    It is given by ::

        𝐳 ∼ 𝒩(Σ(T̃𝛍̃ + K⁻¹𝐦), (T̃ + K⁻¹)⁻¹).
    """

    @property
    def A(self):
        s = self._cov["scale"]
        d = self._cov["delta"]
        return 1 / (s * d * self._site.tau + 1)

    def AQ(self):
        from numpy_sugar.linalg import ddot

        if self._AQ_cache is not None:
            return self._AQ_cache

        Q = self._cov["QS"][0][0]
        A = self.A

        self._AQ_cache = ddot(A, Q)
        return self._AQ_cache

    def ATQ(self):
        from numpy_sugar.linalg import ddot

        if self._TAQ_cache is not None:
            return self._TAQ_cache

        self._TAQ_cache = ddot(self._site.tau, self.AQ())
        return self._TAQ_cache

    def QSQtATQLQtA(self):
        from numpy_sugar.linalg import ddot, dotd

        if self._QSQtATQLQtA_cache is not None:
            return self._QSQtATQLQtA_cache

        LQt = self.LQt()
        A = self.A
        Q = self._cov["QS"][0][0]
        LQtA = ddot(LQt, A)
        AQ = self.AQ()
        QS = self.QS()
        T = self._site.tau
        self._QSQtATQLQtA_cache = dotd(QS, dot(dot(ddot(AQ.T, T), Q), LQtA))
        return self._QSQtATQLQtA_cache

    def L(self):
        r"""Cholesky decomposition of :math:`\mathrm B`.

        .. math::

            \mathrm B = \mathrm Q^{\intercal}\tilde{\mathrm{T}}\mathrm Q
                + \mathrm{S}^{-1}
        """
        from numpy_sugar.linalg import ddot, sum2diag

        if self._L_cache is not None:
            return self._L_cache

        s = self._cov["scale"]
        d = self._cov["delta"]
        Q = self._cov["QS"][0][0]
        S = self._cov["QS"][1]

        ddot(self.A * self._site.tau, Q, left=True, out=self._NxR)
        B = dot(Q.T, self._NxR, out=self._RxR)
        B *= 1 - d
        sum2diag(B, 1.0 / S / s, out=B)
        self._L_cache = _cho_factor(B)
        return self._L_cache

    def update(self):
        from numpy_sugar.linalg import ddot, dotd

        self._flush_cache()

        s = self._cov["scale"]
        d = self._cov["delta"]
        Q = self._cov["QS"][0][0]

        A = self.A
        LQt = self.LQt()

        T = self._site.tau
        E = self._site.eta

        AQ = self.AQ()
        QS = self.QS()

        LQtA = ddot(LQt, A)
        D = self.QSQtATQLQtA()

        self.tau[:] = s * (1 - d) * dotd(QS, AQ.T)
        self.tau -= s * (1 - d) * D * (1 - d)
        self.tau += s * d * A
        self.tau -= s * d * dotd(self.ATQ(), LQtA) * (1 - d)
        self.tau **= -1

        v = s * (1 - d) * dot(Q, dot(QS.T, E)) + s * d * E + self._mean

        self.eta[:] = A * v
        self.eta -= dot(AQ, dot(LQtA, T * v)) * (1 - d)
        self.eta *= self.tau



"""
Mean functions.
"""
from ._kron import KronMean
from ._linear import LinearMean
from ._offset import OffsetMean
from ._sum import SumMean

__all__ = ["OffsetMean", "LinearMean", "SumMean", "KronMean"]



import warnings
from functools import lru_cache, reduce

from numpy import asarray, asfortranarray, kron, log, sqrt, tensordot, trace
from numpy.linalg import matrix_rank, slogdet

from glimix_core._util import log2pi, unvec, vec
from glimix_core.cov import Kron2SumCov
from glimix_core.mean import KronMean
from optimix import Function


class RKron2Sum(Function):
    """
    LMM for multiple traits fitted via restricted maximum likelihood.

    This implementation follows the work published in [CA05]_.
    Let n, c, and p be the number of samples, covariates, and traits, respectively.
    The outcome variable Y is a n×p matrix distributed according to::

        vec(Y) ~ N((A ⊗ F) vec(B), K = C₀ ⊗ GGᵀ + C₁ ⊗ I).

    A and F are design matrices of dimensions p×p and n×c provided by the user,
    where F is the usual matrix of covariates commonly used in single-trait models.
    B is a c×p matrix of fixed-effect sizes per trait.
    G is a n×r matrix provided by the user and I is a n×n identity matrices.
    C₀ and C₁ are both symmetric matrices of dimensions p×p, for which C₁ is
    guaranteed by our implementation to be of full rank.
    The parameters of this model are the matrices B, C₀, and C₁.

    For implementation purpose, we make use of the following definitions:

    - M = A ⊗ F
    - H = MᵀK⁻¹M
    - Yₓ = LₓY
    - Yₕ = YₓLₕᵀ
    - Mₓ = LₓF
    - Mₕ = (LₕA) ⊗ Mₓ
    - mₕ = Mₕvec(B)

    where Lₓ and Lₕ are defined in :class:`glimix_core.cov.Kron2SumCov`.

    References
    ----------
    .. [CA05] Casale, F. P., Rakitsch, B., Lippert, C., & Stegle, O. (2015). Efficient
       set tests for the genetic analysis of correlated traits. Nature methods, 12(8),
       755.
    """

    def __init__(self, Y, A, F, G, rank=1):
        """
        Constructor.

        Parameters
        ----------
        Y : (n, p) array_like
            Outcome matrix.
        A : (n, n) array_like
            Trait-by-trait design matrix.
        F : (n, c) array_like
            Covariates design matrix.
        G : (n, r) array_like
            Matrix G from the GGᵀ term.
        rank : optional, int
            Maximum rank of matrix C₀. Defaults to ``1``.
        """
        Y = asfortranarray(Y, float)
        yrank = matrix_rank(Y)
        if Y.shape[1] > yrank:
            warnings.warn(
                f"Y is not full column rank: rank(Y)={yrank}. "
                + "Convergence might be problematic.",
                UserWarning,
            )

        A = asarray(A, float)
        F = asarray(F, float)
        G = asarray(G, float)
        self._Y = Y
        self._cov = Kron2SumCov(G, Y.shape[1], rank)
        self._mean = KronMean(A, F)
        self._cache = {"terms": None}
        self._cov.listen(self._parameters_update)
        composite = [("C0", self._cov.C0), ("C1", self._cov.C1)]
        Function.__init__(self, "Kron2Sum", composite=composite)

    def _parameters_update(self):
        self._cache["terms"] = None

    @property
    @lru_cache(maxsize=None)
    def _GY(self):
        return self._cov.Ge.T @ self._Y

    @property
    @lru_cache(maxsize=None)
    def _GG(self):
        return self._cov.Ge.T @ self._cov.Ge

    @property
    @lru_cache(maxsize=None)
    def _trGG(self):
        from numpy_sugar.linalg import trace2

        return trace2(self._cov.Ge, self._cov.Ge.T)

    @property
    @lru_cache(maxsize=None)
    def _GGGG(self):
        return self._GG @ self._GG

    @property
    @lru_cache(maxsize=None)
    def _GGGY(self):
        return self._GG @ self._GY

    @property
    @lru_cache(maxsize=None)
    def _FF(self):
        return self._mean.F.T @ self._mean.F

    @property
    @lru_cache(maxsize=None)
    def _GF(self):
        return self._cov.Ge.T @ self._mean.F

    @property
    @lru_cache(maxsize=None)
    def _FGGG(self):
        return self._GF.T @ self._GG

    @property
    @lru_cache(maxsize=None)
    def _FGGY(self):
        return self._GF.T @ self._GY

    @property
    @lru_cache(maxsize=None)
    def _FGGF(self):
        return self._GF.T @ self._GF

    @property
    @lru_cache(maxsize=None)
    def _FY(self):
        return self._mean.F.T @ self._Y

    @property
    def _terms(self):
        from scipy.linalg import cho_factor, cho_solve
        from numpy_sugar.linalg import ddot, sum2diag

        if self._cache["terms"] is not None:
            return self._cache["terms"]

        L0 = self._cov.C0.L
        S, U = self._cov.C1.eigh()
        W = ddot(U, 1 / S) @ U.T
        S = 1 / sqrt(S)
        Y = self._Y
        A = self._mean.A

        WL0 = W @ L0
        YW = Y @ W
        WA = W @ A
        L0WA = L0.T @ WA

        Z = kron(L0.T @ WL0, self._GG)
        Z = sum2diag(Z, 1)
        Lz = cho_factor(Z, lower=True)

        # 𝐲ᵀR⁻¹𝐲 = vec(YW)ᵀ𝐲
        yRiy = (YW * self._Y).sum()
        # MᵀR⁻¹M = AᵀWA ⊗ FᵀF
        MRiM = kron(A.T @ WA, self._FF)
        # XᵀR⁻¹𝐲 = vec(GᵀYWL₀)
        XRiy = vec(self._GY @ WL0)
        # XᵀR⁻¹M = (L₀ᵀWA) ⊗ (GᵀF)
        XRiM = kron(L0WA, self._GF)
        # MᵀR⁻¹𝐲 = vec(FᵀYWA)
        MRiy = vec(self._FY @ WA)

        ZiXRiM = cho_solve(Lz, XRiM)
        ZiXRiy = cho_solve(Lz, XRiy)

        yKiy = yRiy - XRiy @ ZiXRiy
        MKiy = MRiy - ZiXRiM.T @ XRiy
        H = MRiM - XRiM.T @ ZiXRiM
        Lh = cho_factor(H)
        b = cho_solve(Lh, MKiy)
        B = unvec(b, (self.ncovariates, -1))
        self._mean.B = B
        XRim = XRiM @ b

        ZiXRim = ZiXRiM @ b
        mRiy = b.T @ MRiy
        mRim = b.T @ MRiM @ b

        self._cache["terms"] = {
            "b": b,
            "Z": Z,
            "B": B,
            "Lz": Lz,
            "S": S,
            "W": W,
            "WA": WA,
            "YW": YW,
            "WL0": WL0,
            "yRiy": yRiy,
            "MRiM": MRiM,
            "XRiy": XRiy,
            "XRiM": XRiM,
            "ZiXRiM": ZiXRiM,
            "ZiXRiy": ZiXRiy,
            "ZiXRim": ZiXRim,
            "MRiy": MRiy,
            "mRim": mRim,
            "mRiy": mRiy,
            "XRim": XRim,
            "yKiy": yKiy,
            "Lh": Lh,
        }
        return self._cache["terms"]

    @property
    def mean(self):
        """
        Mean 𝐦 = (A ⊗ F) vec(B).

        Returns
        -------
        mean : KronMean
        """
        return self._mean

    @property
    def cov(self):
        """
        Covariance K = C₀ ⊗ GGᵀ + C₁ ⊗ I.

        Returns
        -------
        covariance : Kron2SumCov
        """
        return self._cov

    @property
    def nsamples(self):
        """
        Number of samples, n.
        """
        return self._Y.shape[0]

    @property
    def ntraits(self):
        """
        Number of traits, p.
        """
        return self._Y.shape[1]

    @property
    def ncovariates(self):
        """
        Number of covariates, c.
        """
        return self._mean.F.shape[1]

    def value(self):
        """
        Log of the marginal likelihood.
        """
        return self.lml()

    def gradient(self):
        """
        Gradient of the log of the marginal likelihood.
        """
        return self._lml_gradient()

    @property
    @lru_cache(maxsize=None)
    def _logdet_MM(self):
        M = self._mean.AF
        ldet = slogdet(M.T @ M)
        if ldet[0] != 1.0:
            raise ValueError("The determinant of MᵀM should be positive.")
        return ldet[1]

    def lml(self):
        """
        Log of the marginal likelihood.

        Let 𝐲 = vec(Y), M = A⊗F, and H = MᵀK⁻¹M. The restricted log of the marginal
        likelihood is given by [R07]_::

            2⋅log(p(𝐲)) = -(n⋅p - c⋅p) log(2π) + log(｜MᵀM｜) - log(｜K｜) - log(｜H｜)
                - (𝐲-𝐦)ᵀ K⁻¹ (𝐲-𝐦),

        where 𝐦 = M𝛃 for 𝛃 = H⁻¹MᵀK⁻¹𝐲.

        For implementation purpose, let X = (L₀ ⊗ G) and R = (L₁ ⊗ I)(L₁ ⊗ I)ᵀ.
        The covariance can be written as::

            K = XXᵀ + R.

        From the Woodbury matrix identity, we have

            𝐲ᵀK⁻¹𝐲 = 𝐲ᵀR⁻¹𝐲 - 𝐲ᵀR⁻¹XZ⁻¹XᵀR⁻¹𝐲,

        where Z = I + XᵀR⁻¹X. Note that R⁻¹ = (U₁S₁⁻¹U₁ᵀ) ⊗ I and ::

            XᵀR⁻¹𝐲 = (L₀ᵀW ⊗ Gᵀ)𝐲 = vec(GᵀYWL₀),

        where W = U₁S₁⁻¹U₁ᵀ. The term GᵀY can be calculated only once and it will form a
        r×p matrix. We similarly have ::

            XᵀR⁻¹M = (L₀ᵀWA) ⊗ (GᵀF),

        for which GᵀF is pre-computed.

        The log-determinant of the covariance matrix is given by

            log(｜K｜) = log(｜Z｜) - log(｜R⁻¹｜) = log(｜Z｜) - 2·n·log(｜U₁S₁⁻½｜).

        The log of the marginal likelihood can be rewritten as::

            2⋅log(p(𝐲)) = -(n⋅p - c⋅p) log(2π) + log(｜MᵀM｜)
            - log(｜Z｜) + 2·n·log(｜U₁S₁⁻½｜)
            - log(｜MᵀR⁻¹M - MᵀR⁻¹XZ⁻¹XᵀR⁻¹M｜)
            - 𝐲ᵀR⁻¹𝐲 + (𝐲ᵀR⁻¹X)Z⁻¹(XᵀR⁻¹𝐲)
            - 𝐦ᵀR⁻¹𝐦 + (𝐦ᵀR⁻¹X)Z⁻¹(XᵀR⁻¹𝐦)
            + 2𝐲ᵀR⁻¹𝐦 - 2(𝐲ᵀR⁻¹X)Z⁻¹(XᵀR⁻¹𝐦).

        Returns
        -------
        lml : float
            Log of the marginal likelihood.

        References
        ----------
        .. [R07] LaMotte, L. R. (2007). A direct derivation of the REML likelihood
           function. Statistical Papers, 48(2), 321-327.
        """
        np = self.nsamples * self.ntraits
        cp = self.ncovariates * self.ntraits
        terms = self._terms
        S = terms["S"]
        Lz = terms["Lz"]
        yRiy = terms["yRiy"]
        MRiM = terms["MRiM"]
        mRim = terms["mRim"]
        mRiy = terms["mRiy"]
        XRiy = terms["XRiy"]
        XRiM = terms["XRiM"]
        XRim = terms["XRim"]
        ZiXRiM = terms["ZiXRiM"]
        ZiXRim = terms["ZiXRim"]
        ZiXRiy = terms["ZiXRiy"]

        cov_logdet = log(Lz[0].diagonal()).sum() * 2
        cov_logdet -= 2 * log(S).sum() * self.nsamples
        lml = -(np - cp) * log2pi + self._logdet_MM - cov_logdet

        MKiM = MRiM - XRiM.T @ ZiXRiM
        lml -= slogdet(MKiM)[1]

        yKiy = yRiy - XRiy @ ZiXRiy
        mKiy = mRiy - XRim.T @ ZiXRiy
        mKim = mRim - XRim.T @ ZiXRim
        lml += -yKiy - mKim + 2 * mKiy

        return lml / 2

    def _lml_gradient(self):
        """
        Gradient of the log of the marginal likelihood.

        Let 𝐲 = vec(Y), 𝕂 = K⁻¹∂(K)K⁻¹, and H = MᵀK⁻¹M. The gradient is given by::

            2⋅∂log(p(𝐲)) = -tr(K⁻¹∂K) - tr(H⁻¹∂H) + 𝐲ᵀ𝕂𝐲 - 𝐦ᵀ𝕂(2⋅𝐲-𝐦)
                - 2⋅(𝐦-𝐲)ᵀK⁻¹∂(𝐦).

        Observe that

            ∂𝛃 = -H⁻¹(∂H)𝛃 - H⁻¹Mᵀ𝕂𝐲 and ∂H = -Mᵀ𝕂M.

        Let Z = I + XᵀR⁻¹X and 𝓡 = R⁻¹(∂K)R⁻¹. We use Woodbury matrix identity to
        write ::

            𝐲ᵀ𝕂𝐲 = 𝐲ᵀ𝓡𝐲 - 2(𝐲ᵀ𝓡X)Z⁻¹(XᵀR⁻¹𝐲) + (𝐲ᵀR⁻¹X)Z⁻¹(Xᵀ𝓡X)Z⁻¹(XᵀR⁻¹𝐲)
            Mᵀ𝕂M = Mᵀ𝓡M - 2(Mᵀ𝓡X)Z⁻¹(XᵀR⁻¹M) + (MᵀR⁻¹X)Z⁻¹(Xᵀ𝓡X)Z⁻¹(XᵀR⁻¹M)
            Mᵀ𝕂𝐲 = Mᵀ𝓡𝐲 - (MᵀR⁻¹X)Z⁻¹(Xᵀ𝓡𝐲) - (Mᵀ𝓡X)Z⁻¹(XᵀR⁻¹𝐲)
                  + (MᵀR⁻¹X)Z⁻¹(Xᵀ𝓡X)Z⁻¹(XᵀR⁻¹𝐲)
            H⁻¹   = MᵀR⁻¹M - (MᵀR⁻¹X)Z⁻¹(XᵀR⁻¹M),

        where we have used parentheses to separate expressions
        that we will compute separately. For example, we have ::

            𝐲ᵀ𝓡𝐲 = 𝐲ᵀ(U₁S₁⁻¹U₁ᵀ ⊗ I)(∂C₀ ⊗ GGᵀ)(U₁S₁⁻¹U₁ᵀ ⊗ I)𝐲
                  = 𝐲ᵀ(U₁S₁⁻¹U₁ᵀ∂C₀ ⊗ G)(U₁S₁⁻¹U₁ᵀ ⊗ Gᵀ)𝐲
                  = vec(GᵀYU₁S₁⁻¹U₁ᵀ∂C₀)ᵀvec(GᵀYU₁S₁⁻¹U₁ᵀ),

        when the derivative is over the parameters of C₀. Otherwise, we have

            𝐲ᵀ𝓡𝐲 = vec(YU₁S₁⁻¹U₁ᵀ∂C₁)ᵀvec(YU₁S₁⁻¹U₁ᵀ).

        The above equations can be more compactly written as

            𝐲ᵀ𝓡𝐲 = vec(EᵢᵀYW∂Cᵢ)ᵀvec(EᵢᵀYW),

        where W = U₁S₁⁻¹U₁ᵀ, E₀ = G, and E₁ = I. We will now just state the results for
        the other instances of the aBc form, which follow similar derivations::

            Xᵀ𝓡X = (L₀ᵀW∂CᵢWL₀) ⊗ (GᵀEᵢEᵢᵀG)
            Mᵀ𝓡y = (AᵀW∂Cᵢ⊗FᵀEᵢ)vec(EᵢᵀYW) = vec(FᵀEᵢEᵢᵀYW∂CᵢWA)
            Mᵀ𝓡X = AᵀW∂CᵢWL₀ ⊗ FᵀEᵢEᵢᵀG
            Mᵀ𝓡M = AᵀW∂CᵢWA ⊗ FᵀEᵢEᵢᵀF
            Xᵀ𝓡𝐲 = GᵀEᵢEᵢᵀYW∂CᵢWL₀

        From Woodbury matrix identity and Kronecker product properties we have ::

            tr(K⁻¹∂K) = tr[W∂Cᵢ]tr[EᵢEᵢᵀ] - tr[Z⁻¹(Xᵀ𝓡X)]
            tr(H⁻¹∂H) = - tr[(MᵀR⁻¹M)(Mᵀ𝕂M)] + tr[(MᵀR⁻¹X)Z⁻¹(XᵀR⁻¹M)(Mᵀ𝕂M)]

        Note also that ::

            ∂𝛃 = H⁻¹Mᵀ𝕂M𝛃 - H⁻¹Mᵀ𝕂𝐲.

        Returns
        -------
        C0.Lu : ndarray
            Gradient of the log of the marginal likelihood over C₀ parameters.
        C1.Lu : ndarray
            Gradient of the log of the marginal likelihood over C₁ parameters.
        """
        from scipy.linalg import cho_solve

        terms = self._terms
        dC0 = self._cov.C0.gradient()["Lu"]
        dC1 = self._cov.C1.gradient()["Lu"]

        b = terms["b"]
        W = terms["W"]
        Lh = terms["Lh"]
        Lz = terms["Lz"]
        WA = terms["WA"]
        WL0 = terms["WL0"]
        YW = terms["YW"]
        MRiM = terms["MRiM"]
        MRiy = terms["MRiy"]
        XRiM = terms["XRiM"]
        XRiy = terms["XRiy"]
        ZiXRiM = terms["ZiXRiM"]
        ZiXRiy = terms["ZiXRiy"]

        WdC0 = mdot(W, dC0)
        WdC1 = mdot(W, dC1)

        AWdC0 = mdot(WA.T, dC0)
        AWdC1 = mdot(WA.T, dC1)

        # Mᵀ𝓡M
        MR0M = mkron(mdot(AWdC0, WA), self._FGGF)
        MR1M = mkron(mdot(AWdC1, WA), self._FF)

        # Mᵀ𝓡X
        MR0X = mkron(mdot(AWdC0, WL0), self._FGGG)
        MR1X = mkron(mdot(AWdC1, WL0), self._GF.T)

        # Mᵀ𝓡𝐲 = (AᵀW∂Cᵢ⊗FᵀEᵢ)vec(EᵢᵀYW) = vec(FᵀEᵢEᵢᵀYW∂CᵢWA)
        MR0y = vec(mdot(self._FGGY, mdot(WdC0, WA)))
        MR1y = vec(mdot(self._FY, WdC1, WA))

        # Xᵀ𝓡X
        XR0X = mkron(mdot(WL0.T, dC0, WL0), self._GGGG)
        XR1X = mkron(mdot(WL0.T, dC1, WL0), self._GG)

        # Xᵀ𝓡𝐲
        XR0y = vec(mdot(self._GGGY, WdC0, WL0))
        XR1y = vec(mdot(self._GY, WdC1, WL0))

        # 𝐲ᵀ𝓡𝐲 = vec(EᵢᵀYW∂Cᵢ)ᵀvec(EᵢᵀYW)
        yR0y = vec(mdot(self._GY, WdC0)).T @ vec(self._GY @ W)
        yR1y = (YW.T * mdot(self._Y, WdC1).T).T.sum(axis=(0, 1))

        ZiXR0X = cho_solve(Lz, XR0X)
        ZiXR1X = cho_solve(Lz, XR1X)
        ZiXR0y = cho_solve(Lz, XR0y)
        ZiXR1y = cho_solve(Lz, XR1y)

        # Mᵀ𝕂y = Mᵀ𝓡𝐲 - (MᵀR⁻¹X)Z⁻¹(Xᵀ𝓡𝐲) - (Mᵀ𝓡X)Z⁻¹(XᵀR⁻¹𝐲)
        #       + (MᵀR⁻¹X)Z⁻¹(Xᵀ𝓡X)Z⁻¹(XᵀR⁻¹𝐲)
        MK0y = MR0y - mdot(XRiM.T, ZiXR0y) - mdot(MR0X, ZiXRiy)
        MK0y += mdot(XRiM.T, ZiXR0X, ZiXRiy)
        MK1y = MR1y - mdot(XRiM.T, ZiXR1y) - mdot(MR1X, ZiXRiy)
        MK1y += mdot(XRiM.T, ZiXR1X, ZiXRiy)

        # 𝐲ᵀ𝕂𝐲 = 𝐲ᵀ𝓡𝐲 - 2(𝐲ᵀ𝓡X)Z⁻¹(XᵀR⁻¹𝐲) + (𝐲ᵀR⁻¹X)Z⁻¹(Xᵀ𝓡X)Z⁻¹(XᵀR⁻¹𝐲)
        yK0y = yR0y - 2 * XR0y.T @ ZiXRiy + ZiXRiy.T @ mdot(XR0X, ZiXRiy)
        yK1y = yR1y - 2 * XR1y.T @ ZiXRiy + ZiXRiy.T @ mdot(XR1X, ZiXRiy)

        # Mᵀ𝕂M = Mᵀ𝓡M - (Mᵀ𝓡X)Z⁻¹(XᵀR⁻¹M) - (MᵀR⁻¹X)Z⁻¹(Xᵀ𝓡M)
        #       + (MᵀR⁻¹X)Z⁻¹(Xᵀ𝓡X)Z⁻¹(XᵀR⁻¹M)
        MR0XZiXRiM = mdot(MR0X, ZiXRiM)
        MK0M = MR0M - MR0XZiXRiM - MR0XZiXRiM.transpose([1, 0, 2])
        MK0M += mdot(ZiXRiM.T, XR0X, ZiXRiM)
        MR1XZiXRiM = mdot(MR1X, ZiXRiM)
        MK1M = MR1M - MR1XZiXRiM - MR1XZiXRiM.transpose([1, 0, 2])
        MK1M += mdot(ZiXRiM.T, XR1X, ZiXRiM)

        MK0m = mdot(MK0M, b)
        mK0y = b.T @ MK0y
        mK0m = b.T @ MK0m
        MK1m = mdot(MK1M, b)
        mK1y = b.T @ MK1y
        mK1m = b.T @ MK1m
        XRim = XRiM @ b
        MRim = MRiM @ b

        db = {"C0.Lu": cho_solve(Lh, MK0m - MK0y), "C1.Lu": cho_solve(Lh, MK1m - MK1y)}

        grad = {
            "C0.Lu": -trace(WdC0) * self._trGG + trace(ZiXR0X),
            "C1.Lu": -trace(WdC1) * self.nsamples + trace(ZiXR1X),
        }

        grad["C0.Lu"] += cho_solve(Lh, MK0M).diagonal().sum(1)
        grad["C1.Lu"] += cho_solve(Lh, MK1M).diagonal().sum(1)

        mKiM = MRim.T - XRim.T @ ZiXRiM
        yKiM = MRiy.T - XRiy.T @ ZiXRiM

        grad["C0.Lu"] += yK0y - 2 * mK0y + mK0m - 2 * mdot(mKiM, db["C0.Lu"])
        grad["C0.Lu"] += 2 * mdot(yKiM, db["C0.Lu"])
        grad["C1.Lu"] += yK1y - 2 * mK1y + mK1m - 2 * mdot(mKiM, db["C1.Lu"])
        grad["C1.Lu"] += 2 * mdot(yKiM, db["C1.Lu"])

        grad["C0.Lu"] /= 2
        grad["C1.Lu"] /= 2

        return grad

    def fit(self, verbose=True):
        """
        Maximise the marginal likelihood.

        Parameters
        ----------
        verbose : bool, optional
            ``True`` for progress output; ``False`` otherwise.
            Defaults to ``True``.
        """
        self._maximize(verbose=verbose, pgtol=1e-6)


def _dot(a, b):
    r = tensordot(a, b, axes=([min(1, a.ndim - 1)], [0]))
    if a.ndim > b.ndim:
        if r.ndim == 3:
            return r.transpose([0, 2, 1])
        return r
    return r


def mdot(*args):
    return reduce(_dot, args)


def _sum(a):
    return a.sum(axis=(0, 1))


def mkron(a, b):
    from numpy import kron

    if a.ndim == 3:
        return kron(a.transpose([2, 0, 1]), b).transpose([1, 2, 0])
    return kron(a, b)



from numpy import kron, ravel
from numpy.random import RandomState
from numpy.testing import assert_allclose

from glimix_core.mean import KronMean


def test_mean_kron():

    random = RandomState(0)
    # number of trais
    p = 2
    # number of covariates
    c = 3
    # sample size
    n = 4

    A = random.randn(p, p)
    F = random.randn(n, c)
    B = random.randn(p, c)
    # vecB = random.randn(c, p).ravel()

    mean = KronMean(A, F)
    mean.B = B
    assert_allclose(mean.value(), kron(A, F) @ ravel(B, order="F"))
    assert_allclose(mean._check_grad(), 0, atol=1e-5)



from numpy import asarray, clip, dot, exp
from numpy.linalg import solve

from ._mt_core import MTLMMCore
from ._scan import FastScanner


class MTLMM(MTLMMCore):
    r"""Multi-Trait Linear Mixed Models inference via maximum likelihood.

    It perform inference on the model :eq:`mtlmm1`, explained in the
    :ref:`mtlmm-impl` section.

    Parameters
    ----------
    y : array_like
        Outcome.
    X : array_like
        Covariates as a two-dimensional array.
    QS : tuple
        Economic eigendecompositon in form of ``((Q0, Q1), S0)`` of a
        covariance matrix ``K``.

    Examples
    --------
    .. doctest::

        >>> from numpy import array
        >>> from numpy_sugar.linalg import economic_qs_linear
        >>> from glimix_core.lmm import LMM
        >>>
        >>> X = array([[1, 2], [3, -1]], float)
        >>> QS = economic_qs_linear(X)
        >>> covariates = array([[1], [1]])
        >>> y = array([-1, 2], float)
        >>> lmm = LMM(y, covariates, QS)
        >>> lmm.fit(verbose=False)
        >>> print('%.3f' % lmm.lml())
        -3.649

    One can also specify which parameters should be fitted:

    .. doctest::

        >>> from numpy import array
        >>> from numpy_sugar.linalg import economic_qs_linear
        >>> from glimix_core.lmm import LMM
        >>>
        >>> X = array([[1, 2], [3, -1]], float)
        >>> QS = economic_qs_linear(X)
        >>> covariates = array([[1], [1]])
        >>> y = array([-1, 2], float)
        >>> lmm = LMM(y, covariates, QS)
        >>> lmm.fix('delta')
        >>> lmm.fix('scale')
        >>> lmm.delta = 0.5
        >>> lmm.scale = 1
        >>> lmm.fit(verbose=False)
        >>> print('%.3f' % lmm.lml())
        -3.832
        >>> lmm.unfix('delta')
        >>> lmm.fit(verbose=False)
        >>> print('%.3f' % lmm.lml())
        -3.713
    """

    def __init__(self, y, X, QS=None):
        MTLMMCore.__init__(self, y, X, QS)

    def _get_delta(self):
        from numpy_sugar import epsilon

        v = clip(self._variables.get("logistic").value, -20, 20)
        x = 1 / (1 + exp(-v))
        return clip(x, epsilon.tiny, 1 - epsilon.tiny)

    @property
    def beta(self):
        r"""Get or set fixed-effect sizes."""
        return MTLMMCore.beta.fget(self)

    @beta.setter
    def beta(self, beta):
        MTLMMCore.beta.fset(self, beta)

    def copy(self):
        r"""Return a copy of this object.

        This is useful for performing new inference based on the results
        of the copied object, as the new LMM object will start its inference
        from the initial solution found in the copied object.

        Returns
        -------
        :class:`.LMM`
            Copy of this object.
        """
        o = MTLMM.__new__(MTLMM)

        MTLMMCore.__init__(o, self._y, self.X, self._QS)
        o.delta = self.delta
        if self.isfixed("delta"):
            o.fix("delta")

        setattr(o, "_fix_scale", self._fix_scale)
        setattr(o, "_scale", self._scale)
        setattr(o, "_fix_beta", self._fix_beta)
        setattr(o, "_verbose", self._verbose)

        o.set_nodata()
        return o

    def isfixed(self, var_name):
        r"""Return whether a variable it is fixed or not.

        Parameters
        ----------
        var_name : str
            Possible values are ``"delta"`` and ``"scale"``.

        Returns
        -------
        bool
            ``True`` if fixed; ``False`` otherwise.
        """
        if var_name not in ["delta", "scale", "beta"]:
            msg = "Possible values are 'delta', 'scale', and 'beta'."
            raise ValueError(msg)
        if var_name == "delta":
            return super(MTLMM, self)._isfixed("logistic")
        if var_name == "beta":
            return self._fix_beta
        return self._fix_scale

    @property
    def v0(self):
        r"""First variance.

        Returns
        -------
        float
            :math:`s (1 - \delta)`
        """
        return self.scale * (1 - self.delta)

    @property
    def v1(self):
        r"""Second variance.

        Returns
        -------
        float
            :math:`s \delta`
        """
        return self.scale * self.delta

    def fit(self, verbose=True):
        r"""Maximise the marginal likelihood.

        Parameters
        ----------
        verbose : bool, optional
            ``True`` for progress output; ``False`` otherwise.
            Defaults to ``True``.
        """
        self._verbose = verbose
        if not self.isfixed("delta"):
            self._maximize_scalar(verbose=verbose)
        self.delta = self._get_delta()
        self._update_fixed_effects()
        self._verbose = False

    def fix(self, var_name):
        r"""Disable the optimisation of a given variable.

        Parameters
        ----------
        var_name : str
            Possible values are ``"delta"`` and ``"scale"``.
        """
        if var_name not in ["delta", "scale", "beta"]:
            msg = "Possible values are 'delta', 'scale', and 'beta'."
            raise ValueError(msg)

        if var_name == "delta":
            super(MTLMM, self)._fix("logistic")
        elif var_name == "beta":
            self._fix_beta = True
        else:
            if not self._fix_scale:
                self._scale = self.scale
            self._fix_scale = True

    @property
    def fixed_effects_variance(self):
        r"""Variance of the fixed-effects.

        It is defined as the empirical variance of the prior mean.

        Returns
        -------
        :class:`numpy.ndarray`
            Estimated variance of the fixed-effects.
        """
        return self.mean.var()

    def get_fast_scanner(self):
        r"""Return :class:`.FastScanner` for association scan.

        Returns
        -------
        :class:`.FastScanner`
            Instance of a class designed to perform very fast association scan.
        """
        v0 = self.v0
        v1 = self.v1
        QS = (self._QS[0], v0 * self._QS[1])
        return FastScanner(self._y, self.X, QS, v1)

    def lml(self):
        self.delta = self._get_delta()
        return MTLMMCore.lml(self)

    @property
    def mean(self):
        r"""Estimated mean :math:`\mathrm X\boldsymbol\beta`.

        Returns
        -------
        :class:`numpy.ndarray`
            Mean of the prior.
        """
        return MTLMMCore.mean.fget(self)

    @property
    def scale(self):
        if self._fix_scale:
            return self._scale
        return MTLMMCore.scale.fget(self)

    @scale.setter
    def scale(self, v):
        self._scale = v

    def unfix(self, var_name):
        r"""Enable the optimisation of a given variable.

        Parameters
        ----------
        var_name : str
            Possible values are ``"delta"`` and ``"scale"``.
        """
        if var_name not in ["delta", "scale", "beta"]:
            msg = "Possible values are 'delta', 'scale', and 'beta'."
            raise ValueError(msg)
        if var_name == "delta":
            super(MTLMM, self)._unfix("logistic")
        elif var_name == "beta":
            self._fix_beta = False
        else:
            self._fix_scale = False

    def value(self, *_):
        self.delta = self._get_delta()
        return self.lml()

    @property
    def X(self):
        return MTLMMCore.X.fget(self)

    @X.setter
    def X(self, X):
        MTLMMCore.X.fset(self, X)

    def predictive_mean(self, Xstar, ks, kss):
        mstar = self.mean_star(Xstar)
        ks = self.covariance_star(ks)
        m = self.mean
        K = MTLMMCore.covariance(self)
        return mstar + dot(ks, solve(K, self._y - m))

    def predictive_covariance(self, Xstar, ks, kss):
        kss = self.variance_star(kss)
        ks = self.covariance_star(ks)
        K = MTLMMCore.covariance(self)
        ktk = solve(K, ks.T)
        b = []
        for i in range(len(kss)):
            b += [dot(ks[i, :], ktk[:, i])]
        b = asarray(b)
        return kss - b



from functools import lru_cache

from numpy import (
    asarray,
    atleast_2d,
    concatenate,
    empty,
    eye,
    kron,
    log,
    sqrt,
    tensordot,
    zeros_like,
)
from numpy.linalg import eigh

from optimix import Function

from ._free import FreeFormCov
from ._lrfree import LRFreeFormCov
from .._util import format_function, unvec


class Kron2SumCov(Function):
    """
    Implements K =  C₀ ⊗ GGᵀ + C₁ ⊗ I.

    C₀ and C₁ are d×d symmetric matrices. C₀ is a semi-definite positive matrix while C₁
    is a positive definite one. G is a n×m matrix and I is a n×n identity matrix. Let
    M = Uₘ Sₘ Uₘᵀ be the eigen decomposition for any matrix M. The documentation and
    implementation of this class make use of the following definitions:

    - X = GGᵀ = Uₓ Sₓ Uₓᵀ
    - C₁ = U₁ S₁ U₁ᵀ
    - Cₕ = S₁⁻½ U₁ᵀ C₀ U₁ S₁⁻½
    - Cₕ = Uₕ Sₕ Uₕᵀ
    - D = (Sₕ ⊗ Sₓ + Iₕₓ)⁻¹
    - Lₓ = Uₓᵀ
    - Lₕ = Uₕᵀ S₁⁻½ U₁ᵀ
    - L = Lₕ ⊗ Lₓ

    The above definitions allows us to write the inverse of the covariance matrix as::

        K⁻¹ = LᵀDL,

    where D is a diagonal matrix.


    Example
    -------

    .. doctest::

        >>> from numpy import array
        >>> from glimix_core.cov import Kron2SumCov
        >>>
        >>> G = array([[-1.5, 1.0], [-1.5, 1.0], [-1.5, 1.0]])
        >>> Lr = array([[3], [2]], float)
        >>> Ln = array([[1, 0], [2, 1]], float)
        >>>
        >>> cov = Kron2SumCov(G, 2, 1)
        >>> cov.C0.L = Lr
        >>> cov.C1.L = Ln
        >>> print(cov)
        Kron2SumCov(G=..., dim=2, rank=1): Kron2SumCov
          LRFreeFormCov(n=2, m=1): C₀
            L: [[3.]
                [2.]]
          FreeFormCov(dim=2): C₁
            L: [[1. 0.]
                [2. 1.]]
    """

    def __init__(self, G, dim, rank):
        """
        Constructor.

        Parameters
        ----------
        dim : int
            Dimension d for the square matrices C₀ and C₁.
        rank : int
            Maximum rank of the C₁ matrix.
        """

        self._cache = {"LhD": None}
        self._C0 = LRFreeFormCov(dim, rank)
        self._C0.name = "C₀"
        self._C1 = FreeFormCov(dim)
        self._C1.name = "C₁"
        G = atleast_2d(asarray(G, float))
        self._G = G

        self._Sxe = None
        self._Sx = None
        self._Lx = None
        self._LxG = None
        self._diag_LxGGLx = None
        self._Lxe = None
        self._LxGe = None
        self._diag_LxGGLxe = None

        Function.__init__(
            self, "Kron2SumCov", composite=[("C0", self._C0), ("C1", self._C1)]
        )
        self._C0.listen(self._parameters_update)
        self._C1.listen(self._parameters_update)

    def _init_svd(self):
        from scipy.linalg import svd
        from numpy_sugar.linalg import dotd

        if self._Lx is not None:
            return
        G = self._G
        U, S, _ = svd(G, check_finite=False)
        S *= S
        self._Sxe = S
        self._Sx = concatenate((S, [0.0] * (U.shape[0] - S.shape[0])))
        self._Lx = U.T
        self._LxG = self._Lx @ G
        self._diag_LxGGLx = dotd(self._LxG, self._LxG.T)
        self._Lxe = U[:, : S.shape[0]].T
        self._LxGe = self._Lxe @ G
        self._diag_LxGGLxe = dotd(self._LxGe, self._LxGe.T)

    @property
    @lru_cache(maxsize=None)
    def Ge(self):
        """
        Result of US from the SVD decomposition G = USVᵀ.
        """

        from scipy.linalg import svd
        from numpy_sugar.linalg import ddot

        U, S, _ = svd(self._G, full_matrices=False, check_finite=False)
        if U.shape[1] < self._G.shape[1]:
            return ddot(U, S)
        return self._G

    @property
    @lru_cache(maxsize=None)
    def _GG(self):
        return self._G @ self._G.T

    @property
    @lru_cache(maxsize=None)
    def _I(self):
        return eye(self._G.shape[0])

    def _parameters_update(self):
        self._cache["LhD"] = None

    def listen(self, func):
        """
        Listen to parameters change.

        Parameters
        ----------
        func : callable
            Function to be called when a parameter changes.
        """
        self._C0.listen(func)
        self._C1.listen(func)

    @property
    def Lx(self):
        """
        Lₓ.
        """

        self._init_svd()
        return self._Lx

    @property
    @lru_cache(maxsize=None)
    def _X(self):
        return self.G @ self.G.T

    @property
    def _LhD(self):
        """
        Implements Lₕ and D.

        Returns
        -------
        Lh : ndarray
            Uₕᵀ S₁⁻½ U₁ᵀ.
        D : ndarray
            (Sₕ ⊗ Sₓ + Iₕₓ)⁻¹.
        """
        from numpy_sugar.linalg import ddot

        self._init_svd()
        if self._cache["LhD"] is not None:
            return self._cache["LhD"]
        S1, U1 = self.C1.eigh()
        U1S1 = ddot(U1, 1 / sqrt(S1))
        Sh, Uh = eigh(U1S1.T @ self.C0.value() @ U1S1)
        self._cache["LhD"] = {
            "Lh": (U1S1 @ Uh).T,
            "D": 1 / (kron(Sh, self._Sx) + 1),
            "De": 1 / (kron(Sh, self._Sxe) + 1),
        }
        return self._cache["LhD"]

    @property
    def Lh(self):
        """
        Lₕ.
        """
        return self._LhD["Lh"]

    @property
    def D(self):
        """
        (Sₕ ⊗ Sₓ + Iₕₓ)⁻¹.
        """
        return self._LhD["D"]

    @property
    def _De(self):
        return self._LhD["De"]

    @property
    def G(self):
        """
        User-provided matrix G, n×m.
        """
        return self._G

    @property
    def C0(self):
        """
        Semi-definite positive matrix C₀.
        """
        return self._C0

    @property
    def C1(self):
        """
        Definite positive matrix C₁.
        """
        return self._C1

    def value(self):
        """
        Covariance matrix K = C₀ ⊗ GGᵀ + C₁ ⊗ I.

        Returns
        -------
        K : ndarray
            C₀ ⊗ GGᵀ + C₁ ⊗ I.
        """
        C0 = self._C0.value()
        C1 = self._C1.value()
        return kron(C0, self._GG) + kron(C1, self._I)

    def gradient(self):
        """
        Gradient of K.

        Returns
        -------
        C0 : ndarray
            Derivative of C₀ over its parameters.
        C1 : ndarray
            Derivative of C₁ over its parameters.
        """
        self._init_svd()
        C0 = self._C0.gradient()["Lu"].T
        C1 = self._C1.gradient()["Lu"].T
        grad = {"C0.Lu": kron(C0, self._X).T, "C1.Lu": kron(C1, self._I).T}
        return grad

    def gradient_dot(self, v):
        """
        Implements ∂K⋅v.

        Parameters
        ----------
        v : array_like
            Vector from ∂K⋅v.

        Returns
        -------
        C0.Lu : ndarray
            ∂K⋅v, where the gradient is taken over the C₀ parameters.
        C1.Lu : ndarray
            ∂K⋅v, where the gradient is taken over the C₁ parameters.
        """
        self._init_svd()
        V = unvec(v, (self.G.shape[0], -1) + v.shape[1:])
        r = {}

        C = self._C0.gradient()["Lu"]
        r["C0.Lu"] = tensordot(V.T @ self.G @ self.G.T, C, axes=([-2], [0]))
        r["C0.Lu"] = r["C0.Lu"].reshape(V.shape[2:] + (-1,) + (C.shape[-1],), order="F")

        C = self._C1.gradient()["Lu"]
        r["C1.Lu"] = tensordot(V.T, C, axes=([-2], [0]))
        r["C1.Lu"] = r["C1.Lu"].reshape(V.shape[2:] + (-1,) + (C.shape[-1],), order="F")

        return r

    def solve(self, v):
        """
        Implements the product K⁻¹⋅v.

        Parameters
        ----------
        v : array_like
            Array to be multiplied.

        Returns
        -------
        x : ndarray
            Solution x to the equation K⋅x = y.
        """
        from numpy_sugar.linalg import ddot

        self._init_svd()
        L = kron(self.Lh, self.Lx)
        return L.T @ ddot(self.D, L @ v, left=True)

    def logdet(self):
        """
        Implements log|K| = - log|D| + n⋅log|C₁|.

        Returns
        -------
        logdet : float
            Log-determinant of K.
        """
        self._init_svd()
        return -log(self._De).sum() + self.G.shape[0] * self.C1.logdet()

    def logdet_gradient(self):
        """
        Implements ∂log|K| = Tr[K⁻¹∂K].

        It can be shown that::

            ∂log|K| = diag(D)ᵀdiag(L(∂K)Lᵀ) = diag(D)ᵀ(diag(Lₕ∂C₀Lₕᵀ)⊗diag(LₓGGᵀLₓᵀ)),

        when the derivative is over the parameters of C₀. Similarly,

            ∂log|K| = diag(D)ᵀdiag(L(∂K)Lᵀ) = diag(D)ᵀ(diag(Lₕ∂C₁Lₕᵀ)⊗diag(I)),

        over the parameters of C₁.

        Returns
        -------
        C0 : ndarray
            Derivative of C₀ over its parameters.
        C1 : ndarray
            Derivative of C₁ over its parameters.
        """
        from numpy_sugar.linalg import dotd

        self._init_svd()

        dC0 = self._C0.gradient()["Lu"]
        grad_C0 = zeros_like(self._C0.Lu)
        for i in range(self._C0.Lu.shape[0]):
            t = kron(dotd(self.Lh, dC0[..., i] @ self.Lh.T), self._diag_LxGGLxe)
            grad_C0[i] = (self._De * t).sum()

        dC1 = self._C1.gradient()["Lu"]
        grad_C1 = zeros_like(self._C1.Lu)
        p = self._Sxe.shape[0]
        np = self._G.shape[0] - p
        for i in range(self._C1.Lu.shape[0]):
            t = (dotd(self.Lh, dC1[..., i] @ self.Lh.T) * np).sum()
            t1 = kron(dotd(self.Lh, dC1[..., i] @ self.Lh.T), eye(p))
            t += (self._De * t1).sum()
            grad_C1[i] = t

        return {"C0.Lu": grad_C0, "C1.Lu": grad_C1}

    def LdKL_dot(self, v, v1=None):
        """
        Implements L(∂K)Lᵀv.

        The array v can have one or two dimensions and the first dimension has to have
        size n⋅p.

        Let vec(V) = v. We have

            L(∂K)Lᵀ⋅v = ((Lₕ∂C₀Lₕᵀ) ⊗ (LₓGGᵀLₓᵀ))vec(V) = vec(LₓGGᵀLₓᵀVLₕ∂C₀Lₕᵀ),

        when the derivative is over the parameters of C₀. Similarly,

            L(∂K)Lᵀv = ((Lₕ∂C₁Lₕᵀ) ⊗ (LₓLₓᵀ))vec(V) = vec(LₓLₓᵀVLₕ∂C₁Lₕᵀ),

        over the parameters of C₁.
        """
        self._init_svd()

        def dot(a, b):
            r = tensordot(a, b, axes=([1], [0]))
            if a.ndim > b.ndim:
                return r.transpose([0, 2, 1])
            return r

        Lh = self.Lh
        V = unvec(v, (self.Lx.shape[0], -1) + v.shape[1:])
        LdKL_dot = {
            "C0.Lu": empty((v.shape[0],) + v.shape[1:] + (self._C0.Lu.shape[0],)),
            "C1.Lu": empty((v.shape[0],) + v.shape[1:] + (self._C1.Lu.shape[0],)),
        }

        dC0 = self._C0.gradient()["Lu"]
        for i in range(self._C0.Lu.shape[0]):
            t = dot(self._LxG, dot(self._LxG.T, dot(V, Lh @ dC0[..., i] @ Lh.T)))
            LdKL_dot["C0.Lu"][..., i] = t.reshape((-1,) + t.shape[2:], order="F")

        dC1 = self._C1.gradient()["Lu"]
        for i in range(self._C1.Lu.shape[0]):
            t = dot(V, Lh @ dC1[..., i] @ Lh.T)
            LdKL_dot["C1.Lu"][..., i] = t.reshape((-1,) + t.shape[2:], order="F")

        return LdKL_dot

    def __str__(self):
        dim = self._C0.L.shape[0]
        rank = self._C0.L.shape[1]
        msg0 = format_function(self, {"G": "...", "dim": dim, "rank": rank})
        msg1 = str(self._C0) + "\n" + str(self._C1)
        msg1 = "  " + "\n  ".join(msg1.split("\n"))
        return (msg0 + msg1).rstrip()



from numpy import add

from optimix import Function


class SumCov(Function):
    """
    Sum of covariance functions, K = K₀ + K₁ + ⋯.

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import LinearCov, SumCov
        >>> from numpy.random import RandomState
        >>>
        >>> random = RandomState(0)
        >>> cov_left = LinearCov(random.randn(4, 20))
        >>> cov_right = LinearCov(random.randn(4, 15))
        >>> cov_left.scale = 0.5
        >>>
        >>> cov = SumCov([cov_left, cov_right])
        >>> cov_left.name = "A"
        >>> cov_right.name = "B"
        >>> cov.name = "A+B"
        >>> print(cov)
        SumCov(covariances=...): A+B
          LinearCov(): A
            scale: 0.5
          LinearCov(): B
            scale: 1.0
    """

    def __init__(self, covariances):
        """
        Constructor.

        Parameters
        ----------
        covariances : list
            List of covariance functions.
        """
        self._covariances = [c for c in covariances]
        Function.__init__(self, "SumCov", composite=self._covariances)

    def value(self):
        r"""
        Sum of covariance matrices.

        Returns
        -------
        K : ndarray
            K₀ + K₁ + ⋯
        """
        return add.reduce([cov.value() for cov in self._covariances])

    def gradient(self):
        """
        Sum of covariance function derivatives.

        Returns
        -------
        dict
            ∂K₀ + ∂K₁ + ⋯
        """
        grad = {}
        for i, f in enumerate(self._covariances):
            for varname, g in f.gradient().items():
                grad[f"{self._name}[{i}].{varname}"] = g
        return grad

    def __str__(self):
        tname = type(self).__name__
        msg = "{}(covariances=...)".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        for c in self._covariances:
            spl = str(c).split("\n")
            msg = msg + "\n" + "\n".join(["  " + s for s in spl])
        return msg



from numpy import exp, eye, log

from optimix import Function, Scalar

from .._util import format_function


class EyeCov(Function):
    """
    Identity covariance function, K = s·I.

    The parameter s is the scale of the matrix.

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import EyeCov
        >>>
        >>> cov = EyeCov(2)
        >>> cov.scale = 2.5
        >>> print(cov.value())
        [[2.5 0. ]
         [0.  2.5]]
        >>> g = cov.gradient()
        >>> print(g['logscale'])
        [[2.5 0. ]
         [0.  2.5]]
        >>> cov.name = "I"
        >>> print(cov)
        EyeCov(dim=2): I
          scale: 2.5

    Parameters
    ----------
    dim : int
        Matrix dimension, d.
    """

    def __init__(self, dim):
        """
        Constructor.

        Parameters
        ----------
        dim : int
            Matrix dimension, d.
        """
        self._dim = dim
        self._I = eye(dim)
        self._logscale = Scalar(0.0)
        Function.__init__(self, "EyeCov", logscale=self._logscale)
        self._logscale.bounds = (-20.0, +10)

    @property
    def scale(self):
        """
        Scale parameter.
        """
        return exp(self._logscale)

    @scale.setter
    def scale(self, scale):
        from numpy_sugar import epsilon

        scale = max(scale, epsilon.tiny)
        self._logscale.value = log(scale)

    @property
    def dim(self):
        """
        Dimension of the matrix, d.

        It corresponds to the number of rows and to the number of columns.
        """
        return self._I.shape[0]

    def value(self):
        """
        Covariance matrix.

        Returns
        -------
        K : ndarray
            s⋅I, for scale s and a d×d identity matrix I.
        """
        return self.scale * self._I

    def gradient(self):
        """
        Derivative of the covariance matrix over log(s), s⋅I.

        Returns
        -------
        logscale : ndarray
            s⋅I, for scale s and a d×d identity matrix I.
        """
        return dict(logscale=self.value())

    def __str__(self):
        return format_function(self, {"dim": self._I.shape[0]}, [("scale", self.scale)])
